/*!
\file client.c
\author Robert Sperling <sperling@small-window-manager.de>
\version 1.3.6
\brief functions to manipulate clients and windows

\verbatim
 ***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************
\endverbatim 
*/

/***************************************************************************
 * - client->y|height show the childWindows coordinates                    *
 *   but y has to be set to TITLEBARHEIGHT within parent window            *
 ***************************************************************************/

// -sendConfig--------------------------------------------------------------
//! updates the clients parameters
/*!
\param client	the client to be updated

\verbatim
   v1.3.0-pre1: initial release
   v1.3.6:      added SubstructureNotifyMask to fix a problem with xemacs
                thanks to Laurent Thibaud,
		checking for client in CHECKCLIENTS mode
\endverbatim
*/

void sendConfig(Client *client)
{ XConfigureEvent cevent;

#ifdef TRACE
  trace(1,"sendConfig");
#endif//TRACE

#ifdef CHECKCLIENTS 
  if (client)
  {
#endif//CHECKCLIENTS

#ifdef VERBOSE
    if (client->name)
      swmerr ("win:(%#lx),par:(%#lx),client: %s",
          client->window, client->parent, client->name);
#endif//VERBOSE

    cevent.type 		   = ConfigureNotify; 
    cevent.event 		   = client->window;
    cevent.window 	   	   = client->window;
    cevent.x 		   	   = client->x;
    cevent.y 		   	   = client->y;
    cevent.width 		   = client->width;
    cevent.height 	   	   = client->height;
    cevent.border_width      	   = 0;
    cevent.above 		   = None;
    cevent.override_redirect = 0;
    XSendEvent(display, client->window, False, 
               StructureNotifyMask | SubstructureNotifyMask, 
               (XEvent *)&cevent);
#ifdef CHECKCLIENTS
  }//if client
#ifdef ERRORMSG
  else
  { swmerr ("ERROR: sendConfig: client not valid");
#ifdef DEBUGPARANOID
    exit(-1);
#endif//DEBUGPARANOID  
  }//if !client
#endif//ERRORMSG
#endif//CHECKCLIENTS  

#ifdef TRACE
  trace(-1,"sendConfig");
#endif//TRACE
}//sendConfig


// -redrawParent------------------------------------------------------------
//! redraws the windows titlebar
/*!
\param client	the client which frame should be redrawn

\verbatim
   v1.3.0-pre1: initial release
   v1.3.0-pre2: updated to NOBORDER
   v1.3.0-pre3: updated pixmap support
   v1.3.0-pre4: used stringGC fir border, now fixed
   v1.3.1:      eliminated XClearWindow in MINIMAL mode, eliminated NOBORDER
   v1.3.3:      now only redrawing if client is on current desktop
   v1.3.6:	removed window offsets
\endverbatim
*/
void redrawParent(Client *client)
{ 
#ifndef PIXMAPS
  unsigned int x;
#ifndef MINIMAL
  unsigned int y=TITLEBARHEIGHT/2;
#endif//MINIMAL  
#endif//PIXMAP  

#ifdef SHOWLAYERS
  char stringDummy[512];
#endif//SHOWLAYERS

#ifdef TRACE
  trace(1,"redrawParent");
#endif//TRACE

#ifdef CHECKCLIENTS
  if (client)
  {
#endif//CHECKCLIENTS

#ifdef NAMEHANDLING
#ifdef VERBOSE
    if (client->name) 
      swmerr ("win:(%#lx),par:(%#lx),client: %s",
          client->window, client->parent, client->name);
#endif//VERBOSE
#endif//NAMEHANDLING

#ifdef GNOME
    if (client->desktop==currentDesktop)
    {
#endif//GNOME
    
#ifdef NAMEHANDLING
#ifdef VERBOSE
      if (client->name) 
        swmerr ("redrawing window: %s", client->name);
#endif//VERBOSE
#endif//NAMEHANDLING

#ifndef MINIMAL
      XClearWindow(display,client->parent);
#endif//MINIMAL

#ifdef MWM
      if (client->frameState==FRAMESHOWN)
#endif//MWM
      {

#ifdef VERBOSE
        swmerr ("frame shown");
#endif//VERBOSE

#ifdef PIXMAPS
        XClearWindow(display,client->closebox);
#ifdef MAXIMIZE     
        XClearWindow(display,client->maxbox);
#endif//MAXIMIZE 

#ifdef GNOME     
        XClearWindow(display,client->stickybox); 
#endif//GNOME      

#else//PIXMAPS 
	// closebox 
        x=client->width-TITLEBARHEIGHT;
        XDrawLine(display, client->parent, borderGC, x, 0, x, TITLEBARHEIGHT);

#ifndef MINIMAL    
        x=client->width-TITLEBARHEIGHT/2+1;
        // x-Box  
        XDrawLine(display, client->parent, borderGC,x-4,y-3,x+3,y+4);
        // x-Box  
        XDrawLine(display, client->parent, borderGC,x+3,y-4,x-4,y+3);
#endif//MINIMAL

#ifdef MAXIMIZE
        x=client->width - 3*TITLEBARHEIGHT/2;
        // +-Box left down-line 
        XDrawLine(display, client->parent, borderGC, 
                  client->width-2*TITLEBARHEIGHT, 0,
                  client->width-2*TITLEBARHEIGHT, TITLEBARHEIGHT);

        // +-Box down line of +  
        XDrawLine(display, client->parent, borderGC, x-4, y, x+3,y);
        // +-Box - 
        XDrawLine(display, client->parent, borderGC, x-1, y-3, x-1,y+4);
#endif//MAXIMIZE 

#ifdef GNOME
#ifdef MAXIMIZE
        x=client->width - 5*TITLEBARHEIGHT/2;
#else
        x=client->width - 3*TITLEBARHEIGHT/2;
#endif//MAXIMIZE

        // +-Box left down-line 
        XDrawLine (display, client->parent, borderGC, 
                   x-TITLEBARHEIGHT/2, 0,
                   x-TITLEBARHEIGHT/2, TITLEBARHEIGHT);
        XDrawRectangle(display, client->parent, borderGC, x-4, y-3, 6,6);
        if (client->sticky)
        { XDrawLine (display, client->parent, borderGC, x-4,y-4,x-1,y+1);
          XDrawLine (display, client->parent, borderGC, x-1,y+1,x+5,y-7);
        }
#endif//GNOME
#endif//PIXMAPS 

#ifdef FONTS
#ifdef NAMEHANDLING
        if (client->name)
        { 
#ifdef SHOWLAYERS
          if (client->layer==WIN_LAYER_NORMAL)
#endif//SHOWLAYERS
            XDrawString(display, client->parent, stringGC, 2,
                        sWMFont->ascent+1, client->name, strlen(client->name));
#ifdef SHOWLAYERS
          else
          { sprintf (stringDummy,"(%d)%s",
	             client->layer-WIN_LAYER_NORMAL,client->name);
#ifdef VERBOSE		     
	    swmerr ("Stringdummy is: %s",stringDummy);	     
#endif//VERBOSE
	    
            XDrawString(display, client->parent, stringGC, 2,
                        sWMFont->ascent+1, stringDummy, strlen(stringDummy));
          }
#endif//SHOWLAYERS
        }//if client name
#endif//NAMEHANDLING
#endif//FONTS

      }// end if frameShown
#ifdef MWM
      else
      {
        // Outer Border 
        XDrawRectangle(display, client->parent, borderGC, 0, 0, client->width,
	               client->height);
      }//end else frameshown
#endif//MWM

#ifdef GNOME
    }//desktop==currentdesktop
#endif//GNOME

#ifdef CHECKCLIENTS
  } //if client 
#ifdef ERRORMSG
  else
  { swmerr ("ERROR: redrawParent: client not valid");
#ifdef DEBUGPARANOID
    exit(-1);
#endif//DEBUGPARANOID  
  }//if !client
#endif//ERRORMSG
#endif//CHECKCLIENTS

#ifdef HEAVYSYNC
  XSync(display,false);
#endif//HEAVYSYNC

#ifdef TRACE
  trace(-1,"redrawParent");
#endif//TRACE
}//redrawParent 



#ifdef FAST
// -drawOutline---------------------------------------------------------------
//! redraws the windows border in resize/move mode if FAST
/*!
\param client	client to draw a frame around

\verbatim
   v1.3.0-pre3: initial release
   v1.3.0-pre4: moved rectangle out of the window to avoid expose problems
\endverbatim
*/
void drawOutline(Client *client)
{ 
#ifdef MWM
  if (client->frameState==FRAMESHOWN)
#endif//MWM  
    XDrawRectangle(display, rootWindow, invertGC, 
                   client->x, client->y-TITLEBARHEIGHT, 
                   client->width+1, client->height+TITLEBARHEIGHT+1);
#ifdef MWM
  else
    XDrawRectangle(display, rootWindow, invertGC, 
                   client->x, client->y, 
                   client->width+1, client->height+1);
#endif//MWM

}//drawOutline 
#endif//FAST


// -updateClient--------------------------------------------------------------
//! changes client window according to client record
/*!
\param client		client to update
\param whatToUpdate	UPDATESIZE or not
\todo make updateClient to do smaller updates possible - if size does not change

\verbatim
   v1.3.0-pre1: initial release
   v1.3.0-pre2: updated to NOBORDER option, updated SHADED
   v1.3.0-pre3: updated pixmap support
   v1.3.0-pre4: border messages now DEBUGPARANID
   v1.3.1:      added MAXRESTRICT, ALWAYSMAX
   v1.3.4:      added HEAVYSYNC
   v1.3.6:	removed window offsets,
                checking for client in CHECKCLIENTS mode
\endverbatim
*/
void updateClient(Client *client, int whatToUpdate)
{ int basex, basey;
#ifdef TRACE
  trace (1,"updateClient");
#endif//TRACE

#ifdef CHECKCLIENTS
  if (client)
  {
#endif//CHECKCLIENTS

#ifdef VERBOSE
    if (client->name) swmerr ("updateClientWin:(%#lx),par:(%#lx),client: %s",
                          client->window, client->parent, client->name);
#endif//VERBOSE

    if (whatToUpdate==UPDATESIZE)
    { if (client->size->flags & PResizeInc) 
      { basex = (client->size->flags & PBaseSize) 
                 ? client->size->base_width : (client->size->flags & PMinSize) 
  	         ? client->size->min_width  : 0;
        basey = (client->size->flags & PBaseSize) ? client->size->base_height :
                (client->size->flags & PMinSize) ? client->size->min_height : 0;
                 client->width -= (client->width - basex) % client->size->width_inc;
                 client->height -= (client->height - basey) % client->size->height_inc;
      }//if PResizeInc 

#ifdef ALWAYSMAX
      if (client->size->flags & PMaxSize)
      { // FIXME: Not implemented, yet 
      } else
      { client->x=0;
        client->width=xmax;
#ifdef MWM
        if (client->frameState==FRAMESHOWN)
	{ client->height=ymax-TITLEBARHEIGHT;
          client->y=TITLEBARHEIGHT;
        }//FRAMESHOWN
        else
#endif//MWM	  
	{ client->y=0;
	  client->height=ymax;
	} 
	
      }// elseif MaxSize 
#endif//ALWAYSMAX	       

#ifdef ALWAYSCENTER
      client->x=xmax/2-client->width/2;
      client->y=ymax/2-client->height/2;
#endif//ALWAYSCENTER

#ifdef MAXRESTRICT
      if (client->x < 0)    client->x=0;
      if (client->y < 0)    client->x=TITLEBARHEIGHT;
      if (client->x > xmax) client->x=xmax-TITLEBARHEIGHT;
      if (client->y > ymax) client->y=ymax-TITLEBARHEIGHT;    

      if (client->width+client->x>xmax && (client->x<xmax)) 
        client->width=xmax-client->x;
      if (client->height+client->y>ymax && (client->y<ymax)) 
        client->height=ymax-client->y;
#endif//MAXRESTRICT

#ifdef MWM
      if (client->frameState!=FRAMESHOWN)
      {
#ifdef VERBOSE
        swmerr ("updated: frame not shown, border");
#endif//VERBOSE
        XMoveResizeWindow(display, client->parent, 
                          client->x, client->y, 
                          client->width, client->height);
        XMoveResizeWindow(display, client->window, 0, 0,
                          client->width, client->height);
 			
      }//!FRAMESHOWN				 
      else
#endif//MWM    
      {
#ifdef SHADE      
        if (client->shaded)
          XMoveResizeWindow(display, client->parent, 
  	                    client->x,client->y-TITLEBARHEIGHT, 
			    client->width, TITLEBARHEIGHT);
        else
#endif//SHADE
          XMoveResizeWindow(display, client->parent, 
	                    client->x, client->y-TITLEBARHEIGHT, 
	  		    client->width, client->height+TITLEBARHEIGHT);

#ifdef PIXMAPS
        XMoveResizeWindow(display, client->closebox, 
                          client->width-TITLEBARHEIGHT, 0, 
	  	  	  TITLEBARHEIGHT,TITLEBARHEIGHT);
#ifdef MAXIMIZE
        XMoveResizeWindow(display, client->maxbox, 
                          client->width-2*TITLEBARHEIGHT,0, 
	  	  	  TITLEBARHEIGHT,TITLEBARHEIGHT);
#endif//MAXIMIZE				 

#ifdef GNOME
#ifdef MAXIMIZE
        XMoveResizeWindow(display, client->stickybox, 
                          client->width-3*TITLEBARHEIGHT, 0, 
  		 	  TITLEBARHEIGHT,TITLEBARHEIGHT);
#else//MAXIMIZE
        XMoveResizeWindow(display, client->stickybox, 
                          client->width-2*TITLEBARHEIGHT, 0, 
  		  	  TITLEBARHEIGHT,TITLEBARHEIGHT);
#endif//MAXIMIZE				 
#endif//GNOME				 
#endif//PIXMAP
        XMoveResizeWindow(display, client->window, 0, 
                          TITLEBARHEIGHT, client->width, 
			  client->height);
				 
      }//else if frameHidden 
      sendConfig(client);
    }//if UPDATESIZE

#ifdef HEAVYSYNC
    XSync(display,false);
#endif//HEAVYSYNC

#ifdef CHECKCLIENTS
  }//if client
#ifdef ERRORMSG
  else
  { swmerr ("ERROR: updateClient: client not valid");
#ifdef DEBUGPARANOID
    exit(-1);
#endif//DEBUGPARANOID  
  }//if !client
#endif//ERRORMSG
#endif//CHECKCLIENTS
  
#ifdef TRACE
  trace (-1,"updateClient");
#endif//TRACE
}//updateClient


#ifdef MAXIMIZE
// -maximizeClient-------------------------------------------------------
//! resizes the client to full screen
/*! if gnome is enabled, windowState is set to both 
    WINSTATE_MAXIMIZED_horizontal and WINSTATE_MAXIMIZED_vertical
\param client	client to resize

\verbatim
   v1.3.0-pre1: initial release
   v1.3.6:	update for sWMPanel
\endverbatim   
*/
void maximizeClient(Client *client)
{ 
#ifdef TRACE
  trace(1,"maximizeClient");
#endif//TRACE

#ifdef CHECKCLIENTS
  if (client) 
  { 
#endif//CHECKCLIENTS

#ifdef VERBOSE
  if (client->name) swmerr ("win:(%#lx),par:(%#lx),client: %s",
                        client->window,client->parent,client->name);
#endif//VERBOSE

    if (!client->maximized)
    { client->normalX		= client->x;
      client->normalY		= client->y;
      client->normalWidth	= client->width;
      client->normalHeight	= client->height;
      client->maximized		= true;

#ifdef GNOME
      setHint (client->window, xa_windowState, WIN_STATE_MAXIMIZED); 
#endif//GNOME    
    }// end if not maximized

    client->x			= 0;
    client->width		= xmax;
    
#ifdef MWM    
    if (client->frameState!=FRAMESHOWN)
    { client->y			= 0;
#ifdef SWMPANEL
      client->height		= ymax-SWMPANELSIZE;
#else//SWMPANEL
      client->height		= ymax;
#endif//SWMPANEL      
    }
    else
#endif//MWM   
    { client->y			= TITLEBARHEIGHT;
#ifdef SWMPANEL
      client->height		= ymax-TITLEBARHEIGHT-SWMPANELSIZE;
#else//SWMPANEL
      client->height		= ymax;
#endif//SWMPANEL      

    } //else !FRAMESHOWN   

    if (client->size->flags & PMaxSize) 
    { if (client->width > client->size->max_width) 
        client->width = client->size->max_width;
      if (client->height > client->size->max_height) 
        client->height = client->size->max_height;
    }//if client size hints

    updateClient(client,UPDATESIZE);

#ifdef CHECKCLIENTS
  }// end if client 
#ifdef ERRORMSG
  else
  { swmerr ("ERROR: maximizeParent: client not valid");
#ifdef DEBUGPARANOID
    exit(-1);
#endif//DEBUGPARANOID
  }//if !client
#endif//ERRORMSG
#endif//CHECKCLIENTS

#ifdef TRACE  
  trace(-1,"maximizeClient");
#endif//TRACE  
}//maximizeClient

// -unMaximizeClient---------------------------------------------------------
//! resizes the client to the orginal size
/*!
\param client	to resize
\todo Testing: now not grabbing server to avoid spinlocks

\verbatim
   v1.3.0-pre1: initial release
\endverbatim
*/
void unmaximizeClient(Client *client)
{ 

#ifdef TRACE
  trace(1,"unmaximizeClient");
#endif//TRACE

#ifdef CHECKCLIENTS
  if (client) 
  {
#endif//CHECKCLIENTS

#ifdef VERBOSE
  if (client->name) swmerr ("win:(%#lx),par:(%#lx),client: %s",
                        client->window,client->parent,client->name);
#endif//VERBOSE

    if (client->maximized)
    { client->x=client->normalX;
      client->y=client->normalY;
      client->width=client->normalWidth;
      client->height=client->normalHeight;
      client->maximized=false;  
//      XGrabServer(display);
      updateClient(client,UPDATESIZE);
      sendConfig(client);
//      XUngrabServer(display);
#ifdef GNOME
// delete Property WIN_STATE_MAXIMIZED
      deleteHint (client->window, xa_windowState, WIN_STATE_MAXIMIZED); 
#endif//GNOME    
    }//if maximized 

#ifdef CHECKCLIENTS
  }//if client 
#ifdef ERRORMSG
  else
  { swmerr ("ERROR: unmaximizeClient: client not valid");
#ifdef DEBUGPARANOID
    exit(-1);
#endif//DEBUGPARANOID  
  }//if !client
#endif//ERRORMSG
#endif//CHECKCLIENTS

#ifdef TRACE  
  trace(-1,"unmaximizeClient");
#endif//TRACE  
}//unmaximizeClient 
#endif//MAXIMIZE


#ifdef SHADE
// -toggleClientShaded--------------------------------------------------------
//! shades the window up/down
/*! ignoring any action if the client has no frame
\param client client to shade up or down

\verbatim
   v1.3.0-pre1: initial release, shading is done without updateClient
                unshading uses updateClient
   v1.3.0-pre2: fixed some cosmetics, now using fully updateClient
   v1.3.0-pre3: no change
   v1.3.0-pre4: added XSync after updateClient to ensure proper display
   v1.3.4:      XSyncing now only if HEAVYSYNC is enabled
                sendConfig(client); after updateClient not nessasary anymore
   v1.3.6:      now checking client only in mode CHECKCLIENTS		
\endverbatim
*/
void toggleClientShaded(Client *client) 
{ 
#ifdef TRACE
  trace(1,"toggleClientShaded");
#endif//TRACE

#ifdef CHECKCLIENTS
  if (client) 
  {
#endif//CHECKCLIENTS

#ifdef VERBOSE
  if (client->name) swmerr ("win:(%#lx),par:(%#lx),client: %s",
                        client->window,client->parent,client->name);
#endif//VERBOSE

#ifdef MWM	  
    if (client->frameState==FRAMESHOWN) 
#endif//MWM	    
    {
#ifdef GNOME
      toggleHint(client->window, 
                 xa_windowState, 
		 WIN_STATE_SHADED); 
#endif//GNOME

      if (!client->shaded) 
        client->shaded = true;
      else 
        client->shaded = false;
      
      updateClient(client,UPDATESIZE);

#ifdef HEAVYSYNC
      XSync (display,false);
#endif//HEAVYSYNC
    }//if frameHidden 

#ifdef CHECKCLIENTS
  }//if client
#ifdef ERRORMSG
  else
  { swmerr ("ERROR: toggleClientShaded: client not valid");
#ifdef DEBUGPARANOID
    exit(-1);
#endif//DEBUGPARANOID  
  }//if !client
#endif//ERRORMSG
#endif//CHECKCLIENTS

#ifdef TRACE
  trace(-1,"toggleClientShaded");
#endif//TRACE
}//toggleClientShaded 
#endif//SHADE 

// ***************************************************************************
#ifdef MWM
//  toggleFrameHidden
//! hides/shows the titlebar
/*!
\param client client to frame/unframe

\verbatim
   v1.3.0-pre1: initial release
   v1.3.0-pre2: updated to FRAMELESS
   v1.3.0-pre3: updated pixmap support
   v1.3.4:	HEAVYSYNC included
                sendConfig(client); after updateClient not needed anymore
\endverbatim
*/
void toggleFrameHidden(Client *client) 
{ 
#ifdef TRACE
  trace(1,"toggleFrameHidden");
#endif//TRACE

#ifdef CHECKCLIENTS  
  if (client) 
  {
#endif//CHECKCLIENTS

#ifdef VERBOSE
  if (client->name) swmerr ("win:(%#lx),par:(%#lx),client: %s",
                        client->window, client->parent, client->name);
#endif//VERBOSE

#ifdef SHADE
    if ((!client->shaded) && (client->frameState!=FRAMELESS))
#endif//SHADE
    {
      if (client->frameState==FRAMESHOWN) 
      { client->frameState = FRAMEHIDDEN;
        changeGravity(client,-1);

#ifdef PIXMAPS
	XUnmapWindow (display,client->closebox);
#ifdef MAXIMIZE
	XUnmapWindow (display,client->maxbox);
#endif//MAXIMIZE
#ifdef GNOME
	XUnmapWindow (display,client->stickybox);
#endif//GNOME	
#endif//PIXMAPS 

#ifdef HEAVYSYNC
	XSync(display,false);	
#endif//HEAVYSYNC

        updateClient(client,UPDATESIZE);
	
      }//!frameHidden 
      else 
      { client->frameState = FRAMESHOWN;
        changeGravity(client,1);
#ifdef PIXMAPS
	XMapWindow (display,client->closebox);
#ifdef MAXIMIZE
	XMapWindow (display,client->maxbox);
#endif//MAXIMIZE	
#ifdef GNOME
	XMapWindow (display,client->stickybox);
#endif//GNOME 
#endif//PIXMAPS 
        updateClient(client,UPDATESIZE);
      }//if not frameHidden
      sendConfig(client);
    }//if !shaded

#ifdef CHECKCLIENTS
  }//if client 
#ifdef ERRORMSG
  else
  { swmerr ("ERROR: redrawParent: client not valid");
#ifdef DEBUGPARANOID
    exit(-1);
#endif//DEBUGPARANOID  
  }//if !client
#endif//ERRORMSG
#endif//CHECKCLIENTS

#ifdef TRACE  
  trace(-1,"toggleFrameHidden");
#endif//TRACE
}//toggleFrameHidden 
#endif//MWM

// -hideClient--------------------------------------------------------------
//! hides the client by unmapping its windows
/*!
\param client client to hide
\verbatim
   v1.3.0-pre1: initial release
   v1.3.0-pre3: updated pixmap support
   v1.3.6-pre4: unmapping now all window components
\endverbatim
*/
void hideClient(Client *client)
{ 
#ifdef TRACE
   trace(1,"hideClient");
#endif//TRACE


#ifdef CHECKCLIENTS       
   if (client) 
   {
#endif//CHECKCLIENTS
#ifdef VERBOSE
       if (client->name) swmerr ("win:(%#lx),par:(%#lx),client: %s",
           client->window, client->parent, client->name);
#endif//VERBOSE

     client->ignoreUnmap++;
     XUnmapWindow(display, client->parent);

     //We should now tell the client that it is now unmapped
     XUnmapWindow(display, client->window);    

#ifdef MWM
     if (client->frameState==FRAMESHOWN)
#endif//MWM
     {
#ifdef PIXMAP
       XUnMapWindow(display, client->closebox);
#ifdef MAXIMIZE    
       XUnMapWindow(display, client->maxbox);
#endif//MAXIMIZE
#ifdef GNOME    
       XUnMapWindow(display, client->stickybox);    
#endif//GNOME   
#endif//PIXMAPS
     }//if frameshown

     setValuedHint (client->window, xa_wm_state, IconicState);
#ifdef CHECKCLIENTS
   }//if client
#ifdef ERRORMSG
  else
  { swmerr ("ERROR: hideClient: client not valid");
#ifdef DEBUGPARANOID
    exit(-1);
#endif//DEBUGPARANOID  
  }//if !client
#endif//ERRORMSG
#endif//CHECKCLIENTS
#ifdef TRACE   
   trace(-1,"hideClient");
#endif//TRACE  
}//hideClient 

// -unhideClient------------------------------------------------------------
//! shown the client again after it was unmapped
/*! xm_state is set to Normal
\param client client to show again
\verbatim
   v1.3.0-pre1: initial release
   v1.3.0-pre2: no change
   v1.3.0-pre3: updated pixmap support
   v1.3.0-pre4: moved mapping of buttons to this function
                removed silly typo, showing frame all the time
   v1.3.1:      hmm, silly always mapped parentwin instead of boxes
\endverbatim
*/
void unhideClient(Client *client)
{ 
#ifdef TRACE
  trace(1,"unhideClient");
#endif//TRACE

#ifdef CHECKCLIENTS
  if (client) 
  {
#endif//CHECKCLIENTS  
  
#ifdef VERBOSE  
  if (client->name) swmerr ("win:(%#lx),par:(%#lx),client: %s",
                        client->window, client->parent, client->name);
#endif//VERBSOE

    // tell the client that it should show up again
    setValuedHint (client->window, xa_wm_state, NormalState);
    XMapWindow(display, client->window);
    
#ifdef MWM
    if (client->frameState==FRAMESHOWN)
#endif//MWM
    {
#ifdef PIXMAP
      XMapWindow(display, client->closebox);
#ifdef MAXIMIZE    
      XMapWindow(display, client->maxbox);
#endif//MAXIMIZE
#ifdef GNOME    
      XMapWindow(display, client->stickybox);    
#endif//GNOME   
#endif//PIXMAPS
    }//frameshown
    XMapWindow(display, client->parent);
#ifdef CHECKCLIENTS
  }//if client
#ifdef ERRORMSG
  else
  { swmerr ("ERROR: unhideClient: client not valid");
#ifdef DEBUGPARANOID
    exit(-1);
#endif//DEBUGPARANOID    
  }//if !client
#endif//ERRORMSG
#endif//CHECKCLIENTS  

#ifdef TRACE
  trace (-1,"unhideClient");
#endif//TRACE
}//unhideClient

// ***************************************************************************
//  addClientEntry
//! adds the new window to the client list
/*!
\returns client entry just added to the database
\param givenWin window to add to sWM's database

\verbatim
   v1.3.0-pre6: initial release
   v1.3.6:      removed compile error in malloc
\endverbatim
*/
Client *addClientEntry(Window givenWin)
{ Client *client;
#ifdef NAMEHANDLING
  int status;
#endif//NAMEHANDLING
  
#ifdef TRACE
  trace(1,"addClientEntry");
#endif//TRACE

  client = (Client *)malloc(sizeof (Client));

  if (client!=NULL)
  { if ((client->size = XAllocSizeHints())==NULL)
    { return NULL;
#ifdef ERRORMSG
      swmerr ("ERROR: addClientEntry: new client not valid");
#ifdef DEBUGPARANOID
      exit(1);
#endif//DEBUGPARANOID    
#endif//ERRORMSG
    }//if no memory available for size hints 
    //putting it into the queue 
    client -> next 	= headClient;
    headClient 		= client;
#ifndef MINIMAL
    client->destroyed	= false;
#endif//MINIMAL

#ifdef GNOME
    clientCounter++;
#endif//GNOME
    client->ignoreUnmap = 0;
#ifdef LAYERS
    client->layer	= WIN_LAYER_NORMAL;
#endif//LAYERS    

#ifdef VERBOSE
    swmerr ("getting name");
#endif//VERBOSE
#ifdef NAMEHANDLING
    status=XFetchName(display, givenWin, &client->name);
    if (!status) 
    {
#ifdef ERRORMSG    
      swmerr ("ERROR FETCHING NAME");
#endif//ERRORMSG      
      client->name=NULL;
    }
#endif//NAMEHANDLING

#ifdef VERBOSE
    if(client->name) swmerr ("name is: %s",client->name);
#endif//VERBOSE

#ifdef MAXIMIZE
    client->maximized	= false;
#endif//MAXIMIZE
#ifdef MWM

    client->frameState= false;
#endif//MWM   

#ifdef TRACE
    trace(-1,"addClientEntry (with client)");
#endif//TRACE
    return client;
  }//if client !=null

#ifdef TRACE
  trace(-1,"addClientEntry (with NULL)");
#endif//TRACE
  return NULL;
}//addClientEntry

// ***************************************************************************
//  frameWindow
//! adds the frame around the window
/*!
\verbatim
   v1.3.0-pre6: initial release
   v1.3.6:      checking for client
\endverbatim
*/
void frameWindow (Client *client)
{ XSetWindowAttributes   setAttr;

#ifdef TRACE
  trace(1,"frameWindow");
#endif//TRACE
 
#ifdef CHECKCLIENTS
  if (client)
  {
#endif//CHECKCLIENTS  

    setAttr.override_redirect 	= True;
    setAttr.background_pixel 	= backgroundColor.pixel;
    setAttr.border_pixel 	= borderColor.pixel; 

// Old behaviour of pre 1.3.x versions, not needed anymore
// setAttr.event_mask = ChildMask | ButtonPressMask | 
//                      ExposureMask | EnterWindowMask |
//			ButtonReleaseMask;

// was TESTING, I think it seems to be stable by now
    setAttr.event_mask 		= ChildMask |
				  ExposureMask |     // repaint overwritten
				                     // parent windows
                                  EnterWindowMask |  // needed to change input
				                     // focus if mouse entered
                                  ButtonPressMask;   // needed to manipulate
                                                     // window with mouse 

    client->parent = XCreateWindow(display, 
                                   rootWindow, 
				   client->x, 
    				   client->y - TITLEBARHEIGHT, 
                                   client->width, 
				   client->height + TITLEBARHEIGHT,
	 	                   0, 
				   DefaultDepth(display, screen), 
				   CopyFromParent, 
				   DefaultVisual(display, screen),
                                   CWOverrideRedirect | CWBackPixel | 
				   CWBorderPixel | CWEventMask,
				   &setAttr);

/* old
    client->parent = XCreateWindow(display, rootWindow, client->x-1, 
    				   client->y - TITLEBARHEIGHT-1, 
                                   client->width+1, 
				   client->height + TITLEBARHEIGHT+1,
	 	                   0, DefaultDepth(display, screen), 
				   CopyFromParent, 
				   DefaultVisual(display, screen),
                                   CWOverrideRedirect | CWBackPixel | 
				   CWBorderPixel | CWEventMask,
				   &setAttr);
*/
#ifndef MINIMAL
    XAddToSaveSet(display, client->window);
    XSetWindowBorderWidth(display, client->window, 0);
#endif//MINIMAL

#ifdef NOBORDER
    XSetWindowBorderWidth(display, client->parent, 0);
#else//NOBORDER
    XSetWindowBorderWidth(display, client->parent, 1);
#endif//NOBORDER

    // reparents the client and sets its initial coordinaten IN 
    // the parents window to 0,0
#ifdef MWM
    if (client->frameState==FRAMEHIDDEN)
      XReparentWindow(display, client->window, client->parent, 0, 0);
    else
#endif//MWM
      XReparentWindow(display, client->window, client->parent, 0, TITLEBARHEIGHT);

#ifdef PIXMAPS
    XSetWindowBackgroundPixmap(display, client->parent, backgroundPix.pixmap); 
    client->closebox = XCreateWindow(display, 
    				     client->parent, 
				     client->width-TITLEBARHEIGHT+1, 
				     1, 
                                     TITLEBARHEIGHT, 
				     TITLEBARHEIGHT-2, 
				     0, 
				     DefaultDepth(display, screen), 
                                     CopyFromParent, 
				     DefaultVisual(display, screen),
                                     CWOverrideRedirect | CWBackPixel |
				     CWBorderPixel | CWEventMask,
				     &setAttr);
#ifdef MAXIMIZE
    client->maxbox = XCreateWindow(display, 
    				   client->parent, 
				   client->width-2*TITLEBARHEIGHT+1, 
				   1, 
                                   TITLEBARHEIGHT-2, 
				   TITLEBARHEIGHT-2,
				   0, 
				   DefaultDepth(display, screen), 
                                   CopyFromParent, 
				   DefaultVisual(display, screen),
                                   CWOverrideRedirect | CWBackPixel | 
				   CWBorderPixel | CWEventMask,
				   &setAttr);
#endif//MAXIMIZE				   
#ifdef GNOME
    client->stickybox = XCreateWindow(display, 
    				      client->parent, 
				      client->width-2*TITLEBARHEIGHT+1, 
				      1, 
                                      TITLEBARHEIGHT-2, 
				      TITLEBARHEIGHT-2, 
				      0, 
				      DefaultDepth(display, screen), 
                                      CopyFromParent, 
				      DefaultVisual(display, screen),
                                      CWOverrideRedirect | CWBackPixel | 
				      CWBorderPixel | CWEventMask,
				      &setAttr);
				      

#endif//GNOME 

#ifdef MWM
    if (client->frameState==FRAMESHOWN)
    {
      XMapWindow(display,client->closebox);
      XSetWindowBackgroundPixmap(display, client->closebox, closeBoxPix.pixmap); 
      XSetWindowBorderWidth(display, client->closebox, 0);
#ifdef MAXIMIZE
      XSetWindowBackgroundPixmap(display, client->maxbox, maxBoxPix.pixmap); 
      XSetWindowBorderWidth(display, client->maxbox, 0);
      XMapWindow(display,client->maxbox);
#endif//MAXIMIZE 
#ifdef GNOME
     XSetWindowBorderWidth(display, client->stickybox, 0);
     XMapWindow(display,client->stickybox);
     if (client->sticky) 
      XSetWindowBackgroundPixmap(display, client->stickybox, stickyBoxPix.pixmap); 
     else 
      XSetWindowBackgroundPixmap(display, client->stickybox, unStickyBoxPix.pixmap); 
#endif//GNOME
    }
#endif//MWM      
#endif//PIXMAPS

#ifdef CHECKCLIENTS
  }//if client
#ifdef ERRORMSG
  else
  { swmerr ("ERROR: frameWindow: client not valid");
#ifdef DEBUGPARANOID
    exit(-1);
#endif//DEBUGPARANOID  
  }//if !client
#endif//ERRORMSG
  
#endif//CHECKCLIENTS  

#ifdef TRACE
  trace(-1,"frameWindow");
#endif//TRACE
}//frameWindow 

// ***************************************************************************
//  createNewClient
//! makes sWM manage a new window
/*! Basic functions: collect all known/useful window parameters and
    initialize client record in list 

\param givenWin window to add to management database


\verbatim
   v1.3.0-pre1: initial release  
   v1.3.0-pre2: updated to NOBORDER, added CENTERNEW   
   v1.3.0-pre3: updated pixmap support
   v1.3.0-pre6: fixed some bugs, compressed a bit
   v1.3.4:	remove sendConfig after updateClient,
   		eliminated focusClient
   v1.3.6:      now showing window-close/sticky/maxbox only when frameState
                is FRAMESHOWN		
\endverbatim		
\todo FOR TESTING: Unmapping other windows too, parent should be enough
*/    
void createNewClient (Window givenWin)
{ Client 		*client;
  XWindowAttributes 	 attr;
  XWMHints 		*hints;
#ifdef MWM
  PropMwmHints		*mwmHints;
  unsigned char		*data;
  long			 itemsRead, itemsLeft;
  int			 realFormat;
  Atom			 realType;
#endif//MWM	  
  long  		 dummy;
  
#ifdef TRACE
  trace(1,"createNewClient");
#endif//TRACE
  
  XSetErrorHandler(expectedXErrorHandler);

  client = addClientEntry(givenWin);

  if (!client)
  { 
#ifdef ERRORMSG
    fprintf(stderr,"ERROR: allocating memory for new client failed (PANIC)\n");
#ifdef DEBUGPARANOID
    exit(-1);
#endif//DEBUGPARANOID
#endif//ERRORMSG  

#ifdef TRACE
    trace(-1,"createNewClient (aborted)");
#endif//TRACE
    return;
  }//if !client

  XGetWindowAttributes(display, givenWin, &attr);
  hints 	     = XGetWMHints(display, givenWin);
  XGetWMNormalHints   (display, givenWin, client->size, &dummy);

  client->window     = givenWin;
  client->x 	     = attr.x;
  client->y 	     = attr.y;
  client->width      = attr.width;
  client->height     = attr.height;

#ifdef MAXIMIZE
  client->normalX    = attr.x;
  client->normalY    = attr.y;
  client->normalWidth= attr.width;
  client->normalHeight= attr.height;
#endif//MAXIMIZE

  if (attr.map_state != IsViewable)
  {
#ifdef CENTERNEW
#ifndef FORCECENTERNEW
    if (!(client->size->flags & (PPosition | USPosition)))
    { if (!scanWindowsRunning && client->x==0 && client->y==0) 
      { 
#endif//FORCECENTERNEW      
        client->x=xmax/2-client->width/2;
        client->y=ymax/2-client->height/2;
#ifndef FORCECENTERNEW
      }	    
    }//endif no position given
#endif//FORCECENTERNEW
#endif//CENTERNEW
  }//if !viewable
      
#ifdef ALWAYSCENTER
  //position? That don't impress me much...
  client->x=xmax/2-client->width/2;
  client->y=ymax/2-client->height/2;
#endif//ALWAYSCENTER

#ifdef LOWCOLOR
  client->cmap 	= attr.colormap;
#endif//LOWCOLOR

// Window attributes 
#ifdef SHADE
#ifdef GNOME
  if (getValuedHint(client->window, xa_windowState) & WIN_STATE_SHADED)
  {
#ifdef VERBOSE
   swmerr ("client is SHADED");
#endif//VERBOSE   
    client->shaded	= true;
  }  
  else
#endif//GNOME    
    client->shaded	= false;
#endif//SHADE    

#ifdef GNOME
  client->desktop	= currentDesktop;   
  setValuedHint 	(givenWin, xa_wm_desktop, client->desktop);
  if (getValuedHint(client->window, xa_windowState)&WIN_STATE_STICKY)
  { client->sticky	= true;
#ifdef VERBOSE
    swmerr ("client is STICKY");
#endif//VERBOSE
  }
  else
    client->sticky	= false;

#ifdef VERBOSE
  swmerr ("getting window hints");
#endif//VERBOSE
  
  client->hints=getValuedHint(client->window, xa_windowHints);
#endif//GNOME 


//MWM compliance 
#ifdef MWM
  if (XGetWindowProperty(display, givenWin, xa_mwm_wm_hints, 0L, 20L, False,
      xa_mwm_wm_hints, &realType, &realFormat, &itemsRead,
      &itemsLeft, (unsigned char **) &data)==Success)
  { mwmHints=(PropMwmHints*)data;
    if (mwmHints)
    { if ((mwmHints->flags&MWM_HINTS_DECORATIONS) 
          && !(mwmHints->decorations&MWM_DECOR_ALL))
      { 
#ifdef VERBOSE
        swmerr ("there are decors");
#endif//VERBOSE
        if (mwmHints->decorations&(MWM_DECOR_BORDER|MWM_DECOR_TITLE))
          client->frameState=FRAMESHOWN;
        else 
        {
#ifdef VERBOSE
	  swmerr ("client is frameless");
#endif//VERBOSE
	  client->frameState=FRAMELESS;
        }//else frameshown 
      }//if decorations 
      XFree(mwmHints);
    }//if mwmHints
  }//if properties read   
#endif//MWM    

    // if the client is already mapped, ignore one unmap event 
    if (attr.map_state == IsViewable)
    { client->ignoreUnmap++;
    }
    else
    { // FIXME: initialize window position
      setValuedHint (client->window, xa_wm_state, NormalState);
      if (hints)
      { if (hints->flags & StateHint)
          setValuedHint(client->window, xa_wm_state, hints->initial_state);
      }//if hints
    }//if not viewable

#ifdef MWM
    if (client->frameState==FRAMESHOWN)
#endif//MEM
      changeGravity(client,1);
     
#ifdef LOWCOLOR 
    XSelectInput(display, givenWin, ColormapChangeMask | EnterWindowMask | 
                                    PropertyChangeMask | FocusChangeMask);
#else//LOWCOLOR
    XSelectInput(display, givenWin, PropertyChangeMask | FocusChangeMask | 
                                    EnterWindowMask);
#endif//LOWCOLOR

// **** reparenting **** 
    frameWindow(client);
    updateClient(client,UPDATESIZE);

//    sendConfig(client); not needed anymore

#ifdef LAYERS
  client->layer=getValuedHint(client->window,xa_windowLayer);
#ifdef VERBOSE
  swmerr ("layer: %d",client->layer);
#endif//VERBOSE 
#endif//LAYERS
    if (attr.map_state == IsViewable) 
    { 
      if (getValuedHint(client->window,xa_wm_state) == IconicState) 
      { client->ignoreUnmap++;      
        XUnmapWindow(display,client->window);
#ifndef SHADEICONS
#ifdef PIXMAPS
        XUnmapWindow(display,client->closebox);
#ifdef MAXIMIZE
        XUnmapWindow(display,client->maxbox);
#endif//MAXIMZE
#ifdef GNOME
        XUnmapWindow(display,client->stickybox);
#endif//GNOME	
#endif//PIXMAPS
#endif//SHADEICONS
	XUnmapWindow(display,client->parent);
      }
      else 
      { XMapWindow (display, client->window);

#ifdef MWM
        if (client->frameState==FRAMESHOWN)
	{
#endif//MWM	

#ifdef PIXMAPS
          XMapWindow(display,client->closebox);
#ifdef MAXIMIZE
          XMapWindow(display,client->maxbox);
#endif//MAXIMZE
#ifdef GNOME
          XMapWindow(display,client->stickybox);
#endif//GNOME	
#endif//PIXMAPS
#ifdef MWM
        }//if FRAMESHOWN
#endif//MWM	
        raiseClient(client);
        setValuedHint(client->window,xa_wm_state,NormalState);
      }//else IconicState 
    }//if IsViewable
    else //if not Viewable but normal State
    { if (getValuedHint(client->window,xa_wm_state)==NormalState)
      { XMapWindow(display,client->window);
#ifdef MWM
        if (client->frameState==FRAMESHOWN)
#endif//MWM
        {
#ifdef PIXMAPS
          XMapWindow(display,client->closebox);
#ifdef MAXIMIZE
          XMapWindow(display,client->maxbox);
#endif//MAXIMZE
#ifdef GNOME
          XMapWindow(display,client->stickybox);
#endif//GNOME	
#endif//PIXMAPS
        }
        raiseClient(client);
      }//if NormalState
    }//else not Viewable

#ifdef FOCUSNEW
#ifdef GNOME
        if (!(client->hints & WIN_HINTS_SKIP_FOCUS))
#endif//GNOME
        { focusClient(client); 
	}	 
#endif//FOCUSNEW       

#ifdef CURSORS
    XDefineCursor(display,client->parent,bodyCursor);
    XDefineCursor(display,client->window,windowCursor);
#endif//CURSORS

#ifdef VERBOSE
#ifdef GNOME
    updateClientList();
#endif//GNOME 
#endif//VERBOSE   

// Doing cleanup 
    if (hints) 
    {
#ifdef VERBOSE
      swmerr ("Xfreeing memory in createNewClient");
#endif//VERBOSE
      XFree(hints);
    }//if hints
 
#ifdef PDA
  maximizeClient(client);
#endif//PDA

#ifdef SWMPANELTASKBAR
  showClients();
#endif//SWMPANELTASKBAR

  XSetErrorHandler(xErrorHandler);
#ifdef TRACE
  trace(-1,"createNewClient");
#endif//TRACE
 
}//createNewClient

// ***************************************************************************
//  raiseClient
//! raises a client
/*!
\param client client to raise


\verbatim
   v1.3.0-pre1: initial release
   v1.3.0-pre2: client now won't be raised using XRaiseWindow if Layers are
                supported
   v1.3.0-pre6: adjusting layer was not enough - now really raising
   v1.3.1:      layer management fixed, using adjustLayer
\endverbatim   
*/
void raiseClient(Client *client) 
{
#ifdef TRACE 
  trace(1,"raiseClient"); 
#endif//TRACE

#ifdef CHECKCLIENTS
  if (client)
  { 
#endif//CHECKCLIENTS
#ifdef VERBOSE  
  if (client->name)
    swmerr ("win:(%#lx),par:(%#lx),client: %s",client->window,client->parent,client->name);
#endif//VERBOSE

    unhideClient(client); 

#ifdef LAYERS
    adjustLayer(client,Above);
#else//LAYERS
    XRaiseWindow(display,client->parent);
#endif//LAYERS

#ifdef GNOME
#ifndef NORAISE
    raisedClient=client;
#endif//NORAISE
#endif//GNOME

#ifdef CHECKCLIENTS
  }//if client
#ifdef ERRORMSG
  else
  { swmerr ("ERROR: raiseClient: client not valid");
#ifdef DEBUGPARANOID
    exit(-1);
#endif//DEBUGPARANOID    
  }//if !client
#endif//ERRORMSG  
#endif//CHECKCLIENTS 

#ifdef TRACE 
  trace(-1,"raiseClient"); 
#endif//TRACE 
}//raiseClient  

// ***************************************************************************
#ifndef MINIMAL
//  lowerClient
//! moves a client to the background
/*!
\verbatim
   v1.3.0-pre2: initial release
\endverbatim
*/
void lowerClient(Client *client) 
{ 
#ifdef TRACE 
  trace(1,"lowerClient"); 
#endif//TRACE

#ifdef CHECKCLIENTS
  if (client)
  {  
#endif//CHECKCLIENTS

#ifdef VERBOSE  
    if (client->name) swmerr ("win:(%#lx),par:(%#lx),client: %s",
        client->window,client->parent,client->name);
#endif//VERBOSE	

    unhideClient(client); 

#ifdef LAYERS 
    adjustLayer(client,False);
#else//LAYERS
    XLowerWindow(display,client->parent);
#endif//LAYERS

#ifdef GNOME
#ifndef NORAISE
// FIXME get raised client
    if (!raisedClient) raisedClient=headClient;
    if (raisedClient==client) 
     raisedClient=getFocusedClient();
#endif//NORAISE     
#endif//GNOME     

#ifdef CHECKCLIENTS
  }//if client
#ifdef ERRORMSG
  else
  { swmerr ("ERROR: lowerClient: client not valid");
#ifdef DEBUGPARANOID
    exit(-1);
#endif//DEBUGPARANOID    
  }//if !client
#endif//ERRORMSG
#endif//CHECKCLIENTS

#ifdef TRACE 
  trace(-1,"lowerClient"); 
#endif//TRACE
}//lowerClient
#endif//MINIMAL 

/***************************************************************************/
//  removeClientEntry
//! removes a clients entry from sWM's list
/*!
\verbatim
   v1.3.0-pre5: initial release
   v1.3.6:      forgot to free client structure, fixed thanks to Ludovic Drolez
\endverbatim   
*/

void removeClientEntry (Client *client)
{ Client *previous;
#ifdef TRACE
  trace(1,"removeClientEntry");
#endif//TRACE

  if(!client)
  {
#ifdef ERRORMSG  
    swmerr ("client not found! Nothing to remove.");
#endif//ERRORMSG    
    return;
  }//if !client

  // this should be watched carefully - the head client may be NULL
  if (headClient == client) 
  { headClient = client->next;
    clientCounter--;
#ifdef VERBOSE
    swmerr ("client==headClient");
#endif//VERBOSE    
  }//if headclient==client
  else
  { // client!=headClient
#ifdef VERBOSE  
    swmerr ("client!=headClient");
#endif//VERBOSE    
    previous=headClient;

    while ((previous->next) && (previous->next!=client))
      previous=previous->next; 

    // now previous->next should be our client
    if (previous->next)
    { if (previous->next==client) 
      { previous->next=client->next;
#ifdef VERBOSE      
        swmerr ("REMOVING client");
#endif//VERBOSE	
        clientCounter--;
        if (client->size) XFree(client->size);
        if (client)       free(client);
      }
      else
      {
#ifdef ERRORMSG      
        fprintf(stderr,"ERROR/PANIC removeClientEntry did'nt find its own client\n");
#ifdef DEBUGPARANOID
        exit(-1);
#endif//DEBUGPARANOID	
#endif//ERRORMSG	
      }//if previous->next	
    }//if client found	
    else
    { 
#ifdef ERRORMSG    
      fprintf(stderr,"previous->next does not exist, client not removed!\n");
#endif//ERRORMSG      
    }
  }//else if headClient=client  

  
#ifdef SWMPANELTASKBAR
  showClients();
#endif//SWMPANELTASKBAR

#ifdef TRACE
  trace(-1,"removeClientEntry");
#endif//TRACE
}//removeClientEntry 

/***************************************************************************/
//  removeParent
//! removes a clients frame carefully
/*!
\verbatim
   v1.3.0-pre5: initial release
   v1.3.2:      in MINIMAL mode, now not using safesets anymore, compressed,
                now using only expectedXErrorHandler
   v1.3.5:      now setting ignoring error handlers active until after 
                updateClient
   v1.3.6:      now ignoring XErrors via expectedXErrorHandler avoiding
                to be called more than once
\endverbatim
\param client  the client that should be unframed/removed
\param mapAgain true, if the client window should be shown again (sWM quits)
*/
void removeParent(Client *client, int mapAgain)
{
#ifdef TRACE
  trace(1,"removeParent");
#endif//TRACE

#ifdef VERBOSE
  swmerr ("before barrier");
  while (removeParentCurrentWindow) 
  { swmerr ("waiting..."); 
  }
  swmerr ("after barrier");
#endif//VERBOSE

  if (client)    
    removeParentCurrentWindow=client->window;
#ifdef ERRORMSG
  else 
  { swmerr ("ERROR: got no client... something is wrong here!");
#ifdef DEBUGPARANOID
    exit(-1);
#endif//DEBUGPARANOID   
  }//if !client    
#endif//ERRORMSG  
  
#ifdef VERBOSE
  if (client && client->name)
  { swmerr ("win:(%#lx),par:(%#lx),client: %s",
        client->window,	client->parent,	client->name);
  }
  swmerr ("after win");
#endif//VERBOSE

  //Now setting it to fully ignore, since removeParent should
  //not be called more than once
  XSetErrorHandler(expectedXErrorHandler);

#ifdef VERBOSE
  swmerr ("now grabbing server");
#endif//VERBOSE
  XGrabServer(display);
  
  if (client->window)
  {
#ifndef MINIMAL  
    XRemoveFromSaveSet(display, client->window);
    XSetWindowBorderWidth(display, client->window, 1);
#endif//MINIMAL

#ifdef MWM
    if (client->frameState==FRAMESHOWN)
#endif//MWM
      changeGravity(client,-1);

    XReparentWindow(display, client->window, rootWindow, client->x,client->y);

    if (mapAgain) 
      XMapWindow (display, client->window);
    else
      setValuedHint (client->window, xa_wm_state, WithdrawnState); 
  }//if window valid

  XDestroyWindow(display, client->parent);

#ifdef PIXMAPS
  if (client->closebox) 
    XDestroyWindow(display, client->closebox);
#ifdef MAXIMIZE    
  if (client->maxbox) 
    XDestroyWindow(display, client->maxbox);
#endif//MAXIMIZE
#ifdef GNOME
  if (client->stickybox)  
    XDestroyWindow(display, client->stickybox);
#endif//GNOME
#endif//PIXMAPS    

#ifdef VERBOSE
  swmerr ("running removeClientEntry");
#endif//VERBOSE  
  removeClientEntry (client);
    
#ifdef VERBOSE
  swmerr ("ungrabbing in removeClient");
#endif//VERBOSE
  XUngrabServer(display);

#ifdef VERBOSE
#ifdef GNOME
  updateClientList();
#endif//GNOME
#endif//VERBOSE
  removeParentCurrentWindow=0;
  XSetErrorHandler(xErrorHandler);
#ifdef TRACE  
   trace(-1,"removeParent");
#endif//TRACE
}//removeParent 

// ***************************************************************************
//  sendExitClient
//! sends the client a message to quit itself nicely
/*! if it doesn`t listen to xa_wm_delete, we try to kill it as nicely as 
    possible.

\verbatim
   v1.2.6 initial release
   v1.3.0-pre1: removed old "ignore" stuff
   v1.3.0-pre2: freeing (forgotten) protocols
                included raise/focus management
   v1.3.0-pre4: inserted some fine tuning for minimal mode
   v1.3.4:      added heavysync
\endverbatim
*/
void sendExitClient(Client *client)
{ int count, maxcount, found = 0;
  Atom *protocols;

#ifdef TRACE  
  trace(1,"sendExitClient");
#endif//TRACE

#ifdef CHECKCLIENTS
  if (client)
  { if (client->name) 
      swmerr ("win:(%#lx),par:(%#lx),client: %s",client->window,client->parent,client->name);
#endif//CHECKCLIENTS

#ifdef GNOME
#ifndef NORAISE
    if (raisedClient==client) raisedClient=client->next;
    if (!raisedClient) raisedClient=headClient;
#endif//NORAISE 
#endif//GNOME 

    if (XGetWMProtocols(display, client->window, &protocols, &maxcount))
    { for (count=0; count<maxcount; count++) 
      { if (protocols[count] == xa_wm_delete) found++;
      }
      if (protocols) XFree(protocols);
    }
    if (found) 
    { 
#ifdef VERBOSE
      swmerr ("sending client message to quit nicely");
#endif//VERBOSE    
      sendXMessage(client->window, xa_wm_protos, xa_wm_delete);
    }
    else 
    { 
#ifdef VERBOSE
    swmerr ("pre-syncing");
#endif//VERBOSE    

#ifdef HEAVYSYNC
    XSync (display,false);
#endif//HEAVYSYNC
#ifdef VERBOSE
      swmerr ("sending XKillClient to kill client");
#endif//VERBOSE    
      XKillClient(display, client->window);

#ifndef MINIMAL
      client->destroyed=true;
#endif//MINIMAL
    }//else if found

#ifdef CHECKCLIENTS
  }// end if(client)
#ifdef ERRORMSG
  else
  { swmerr ("error: trying to delete nonexisting client"); 
#ifdef DEBUGPARANOID
    exit(-1);
#endif//DEBUGPARANOID   
  }
#endif//ERRORMSG  
#endif//CHECKCLIENTS

#ifdef TRACE
  trace(-1,"sendExitClient");
#endif//TRACE
}//sendExitClient

  
// ***************************************************************************
//  moveClient
//! handles window movement 
/*!
\verbatim
   v1.3.0-pre1: initial release
   v1.3.0-pre3: updated to support FAST again, added handledClient 
   v1.3.1:      removed old STARTACTION stuff
\endverbatim   
*/
void moveClient(Client *client)
{
#ifdef TRACE
  trace (1,"moveClient");
#endif//TRACE

#ifdef CHECKCLIENTS
  if (client) 
  {
#endif//CHECKCLIENTS
    pointerMode = POINTERMOVEMODE;

#ifdef CURSORS
    XGrabPointer(display, client->parent, False, 
                 PointerMotionMask | ButtonReleaseMask, GrabModeAsync, 
		 GrabModeAsync, None, moveCursor, CurrentTime); 
#else//CURSORS
    XGrabPointer(display, client->parent, False, 
                 PointerMotionMask | ButtonReleaseMask, GrabModeAsync, 
		 GrabModeAsync, None, None, CurrentTime); 
#endif//CURSORS 

    handledClient=client;

#ifdef FAST
    drawOutline(client);
#endif//FAST    

#ifdef CHECKCLIENTS
  }//if client 
#ifdef ERRORMSG
  else
  { swmerr ("ERROR: moveClient: client not valid");
#ifdef DEBUGPARANOID
    exit(-1);
#endif//DEBUGPARANOID    
  }//if !client
#endif//ERRORMSG
#endif//CHECKCLIENTS

#ifdef TRACE
  trace (-1,"moveClient");
#endif//TRACE
}//moveClient

// -resizeClient-------------------------------------------------------------
//! handles window resizeing - start, resizing, end
/*!
\verbatim
   v1.3.0-pre1: initial release
   v1.3.0-pre3: updated to support FAST again, added handledClient 
   v1.3.0-pre4: added TRACE, added check for valid client
   v1.3.1:      added NORESIZEWARP, removed old STARTACTION stuff
\endverbatim
*/
void resizeClient(Client *client)
{
#ifdef TRACE
  trace (1,"resizeClient");
#endif//TRACE

#ifdef NORESIZEWARP
  if (handledClient_old_y) pointerMode = POINTERBOXRESIZEMODE;
  else
#endif//NORESIZEWARP  
    pointerMode = POINTERRESIZEMODE;

#ifdef CHECKCLIENTS
  if (client)
  {
#endif//CHECKCLIENS

#ifdef CURSORS
    XGrabPointer(display, client->parent, False, 
                 PointerMotionMask | ButtonReleaseMask, GrabModeAsync,   
                 GrabModeAsync, None, resizeCursor, CurrentTime); 
#else//CURSORS
    XGrabPointer(display, client->parent, False, 
                 PointerMotionMask | ButtonReleaseMask, GrabModeAsync,   
  	         GrabModeAsync, None, None, CurrentTime); 
#endif//CURSORS

#ifndef NORESIZEWARP
    XWarpPointer(display,None,client->window,0,0,0,0,
                 client->width+1,client->height+1);
#endif//NORESIZEWARP

#ifdef FAST
    updateClient(client,UPDATESIZE);
    drawOutline(client);
#endif//FAST

    handledClient=client;
#ifdef CHECKCLIENTS
  }//if client
#ifdef ERRORMSG
  else
  { swmerr ("ERROR: resizeClient: client not valid");
#ifdef DEBUGPARANOID
    exit(-1);
#endif//DEBUGPARANOID    
  }//if !client
#endif//ERRORMSG
#endif//CHECKCLIENS

#ifdef TRACE
  trace (-1,"resizeClient");
#endif//TRACE
}//resizeClient
