/*! 
\file eventHandler.c
\version 1.3.6
\brief Handling routines for all events

This file contains all functions to react on X-server events

\verbatim
 ***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *************************************************************************** 
\endverbatim

*/

#ifdef SIGNALS
// *************************************************************************
// Signal handlers
// *************************************************************************

//  signalHandler
//! reacts on signals
/*!
\param signalNr Number of the signal sent
\verbatim  
   v1.3.0-pre4: initial release
\endverbatim

*/  
void signalHandler(int signalNr)
{
#ifdef DEBUG
#ifdef TRACE
  if (signalNr==SIGUSR1)
    dumpClients();
  else if (signalNr==SIGUSR2)
  { int counter=0;
    fprintf(stdout,"SIGUSR traceTree: main");
    for (counter=0; counter<debuglevel+2; counter++)
    { fprintf(stdout,"%s->",traceTree[counter]);
    }
    fprintf(stdout,"end\n");
  } 
  else 
#endif//TRACE
#endif//DEBUG

#ifdef SWMPANELCLOCK
  if (signalNr==SIGALRM)
  { refreshPanelClock();
    setPanelClockTimer(true); //wake up again in one minute
#ifdef VERBOSE
    swmerr ("got clock signal, refreshing panel clock");
#endif//VERBOSE  
  } 
  else
#endif//SWMPANELCLOCK  
    sWMQuitNicely(0);  
}//signalHandler
#endif//SIGNALS

// ***************************************************************************

//  mapRequestHandler
//! reacts, when a client wants to be mapped again or if a new window was created
/*!
\verbatim  
   v1.3.0-pre1: initial release
   v1.3.0-pre3: updated for pixmaps
   v1.3.0-pre4: now showing name of affected client
\endverbatim   
*/  
void mapRequestHandler(XMapRequestEvent *event)
{ Client *client;

#ifdef TRACE
  trace (1,"mapRequestHandler"); 
#endif //TRACE

#ifdef SWMPANEL
  if (event->window==sWMWindow) return;
#endif//SWMPANEL

// ****  DEBUG PART  **** 
#ifdef VERBOSE
  client = getClient(event->window, anyWindow);
  if (client) 
  { if (client->name) 
      swmerr ("window: %s",client->name);
  }
#endif//VERBOSE

// **** FUNCTION PART ****
  client = getClient(event->window, childWindow);

  if (client) unhideClient(client);
  else        createNewClient(event->window);

#ifdef TRACE
  trace (-1,"mapRequestHandler"); 
#endif//TRACE
}//mapRequestHandler



// -unmapEventHandler------------------------------------------------------
//! unmaps a client or destroys it if it has been unmapped to often
/*! If the client was unmapped by sWM, decrement ignoreUnmap,
    if it unmapped/destroyed itself, remove the frame & clear it from list

\verbatim  
   v1.3.0-pre1: initial release
   v1.3.0-pre4: showning name of window
   v1.3.0-pre5: now handles unmap events of clients only 
   v1.3.6:      oops, err was used with old parameter stderr
\endverbatim
*/
void unmapEventHandler(XUnmapEvent *event)
{ Client *client;

#ifdef TRACE
  trace (1,"unmapEventHandler"); 
#endif//TRACE

#ifdef VERBOSE
  swmerr ("unmapEventHandler runs for window, event: %#lx,%#lx\n",
      event->window,event->event);
#endif//VERBOSE

#ifdef SWMPANEL
  if (event->window==sWMWindow) return;
#endif//SWMPANEL

#ifndef FASTSYNC
  XSync(display,false);
#endif//FASTSYNC

//  XGrabServer(display);

// ****  DEBUG - PART  **** 
#ifdef CHECKCLIENTS
  client = getClient(event->window, anyWindow);
  if (client) 
  { 
#ifdef VERBOSE  
    if (client->name) 
      swmerr ("window: %s, unmaps %i",client->name, client->ignoreUnmap);
      
    if (event->window==client->window) 
      swmerr ("unmapEvent for clientWindow");
#endif//VERBOSE      
  }//if client
#ifdef ERRORMSG
  else
  { swmerr ("unmapped window belongs to no client I know");
    swmerr ("window ID is: %#lx",event->window);
  }
#endif//ERRORMSG  
#endif//CHECKCLIENTS

// **** FUNCTION - PART *****
  client = getClient(event->window, childWindow);

  if (client) 
  { if (client->ignoreUnmap>0) client->ignoreUnmap--;
    else                       removeParent(client, false); // don't map again
  }//if client 
#ifdef VERBOSE
    swmerr ("returned from removeParent");
#endif//VERBOSE
  
//  XUngrabServer(display);

#ifdef TRACE
  trace (-1,"unmapEventHandler"); 
#endif // TRACE 
}//unmapEventHandler 



#ifndef MINIMAL
// -clientMessageHandler----------------------------------------------------
//! handles client messages
/*!
\verbatim
   v1.3.0-pre1: initial release
   v1.3.0-pre2: fixed hint management
   v1.3.0-pre4: updated desktop change request
   v1.3.1:      now inconify requests can be customized: NOICONS,SHADEICONS,
                and old hide behaviour
   v1.3.2:      MINIMAL mode now implies NOICONS so clientMessageHandler
                will not be called since it is of no use
\endverbatim		
*/
void clientMessageHandler(XClientMessageEvent *event)
{ Client *client;

#ifdef TRACE
  trace (1,"clientMessageHandler"); 
#endif//TRACE

// ****   DEBUG - PART   ****
#ifdef VERBOSE
  client = getClient(event->window, anyWindow);
  if (client && client->name) swmerr ("window: %s",client->name);
#endif//VERBOSE

// **** FUNCTION - PART **** 
  client = getClient(event->window, childWindow);

  if (client)
  { 
#ifndef NOICONS  
    // I don't allow any client to iconify, it will be hidden 
    // or shaded instead   
    if (event->message_type == xa_wm_change_state &&
        event->format == 32 && event->data.l[0] == IconicState) 
    { 
#ifdef SHADEICONS
        toggleClientShaded(client);
#else // SHADEICONS
// TESTING: Seems I did it too often...
//      client->ignoreUnmap++;
      hideClient(client);
#endif // SHADEICONS
    }	
#endif // NOICONS

#ifdef GNOME
#ifdef SHADE
    // client wants to toggle shade 
    if(event->message_type==xa_windowState && event->format==32 && 
       event->data.l[0]&WIN_STATE_SHADED) 
    { toggleClientShaded(client);
    } 
#endif //SHADE

    // client wants to toggle sticky
    if(event->message_type==xa_windowState && event->format==32 && 
       event->data.l[0]&WIN_STATE_STICKY) 
      toggleClientSticky(client); 

    // client wants to change to another desktop 
    if(event->message_type==xa_wm_desktop && event->format==32 &&
       event->data.l[0]!=currentDesktop)
      moveClientToDesktop (client,event->data.l[0]);
#endif//GNOME    

#ifdef LAYERS 
 // client wants to change its layer 
  if(event->message_type==xa_windowLayer && event->format==32) 
    setLayer(client, event->data.l[0]); 
#endif//LAYERS
 }//if client 
#ifdef GNOME
 else 
 { if (event->window==rootWindow)
   { if (event->message_type == xa_wm_desktop && event->format==32)
       changeDesktop(event->data.l[0]);  
   }
 }//else if client 
#endif//GNOME
 
#ifdef TRACE
  trace (-1,"clientMessageHandler"); 
#endif//TRACE
}//clientMessageHandler 
#endif//NOT MINIMAL

// ***************************************************************************

#ifdef LOWCOLOR
//  colormapChangeHandler
//! keeps track of changing colormaps on focus change
/*!
\verbatim  
   v1.3.0-pre1: initial release
\endverbatim
*/
void colormapChangeHandler(XColormapEvent *event)
{ Client *client;

#ifdef TRACE
  trace (1,"colormapChangeHandler");
#endif

// DEBUG - PART 
#ifdef VERBOSE
  client = getClient(event->window, anyWindow);
  if (client) 
  { if (client->name) swmerr ("window: %s",client->name);
  }
#endif//VERBOSE

// FUNCTION - PART 
  client = getClient(event->window, childWindow);

  if (client)
   if (event->new) 
   { client->cmap = event->colormap;
     XInstallColormap(display, client->cmap);
   }// end if new colormap 

#ifdef TRACE
  trace(-1,"colormapChangeHandler");
#endif

}// end colormapChangeHandler 
#endif // LOWCOLOR 

// -propertyChangeHandler--------------------------------------------------
//! handles property change notifies
/*!
\verbatim   
   v1.3.0-pre1: initial release
   v1.3.0-pre3: removed unnessesary XClearWindow 
   v1.3.1:	added handler for desktop & layers
   v1.3.3:	now caching xa_windowHints
   v1.3.4a:	fixed bug with xa_windowHints in non-GNOME mode
   v1.3.6:	when client resizes itself, maximizes is set to false
\endverbatim   
\todo FOR TESTING: when client gets wm_normal_hints, should size be adjusted?
\todo FOR TESTING: using adjustLayer now
*/
void propertyChangeHandler(XPropertyEvent *event)
{ Client *client;
#ifdef NAMEHANDLING
  int status;
#endif//NAMEHANDLING

#ifdef LAYERS
  int oldLayer;
#endif//LAYERS  

#ifdef TRACE
  trace (1,"propertyChangeHandler"); 
#endif//TRACE

#ifdef VERBOSE
  swmerr ("got changed atom: %s",XGetAtomName(display,event->atom));
#endif//VERBOSE

// DEBUG - PART 
#ifdef VERBOSE
  client = getClient(event->window, anyWindow);
  if (client) 
  { if (client->name) swmerr ("window: %s",client->name);
    else swmerr ("unnamed client");
  }	  
  else
  { swmerr ("propertyEvent: this event is not for sWM");
    return;//a little speedup
  }  
#endif//VERBOSE

// FUNCTION - PART 
  client = getClient(event->window, childWindow);
  if (client) 
  { 
#ifdef VERBOSE
    swmerr ("propertyChange on client window");
#endif//VERBOSE	  
#ifdef GNOME
      if (event->atom == xa_windowHints)
      { client->hints=getValuedHint(client->window,xa_windowHints);
#ifdef VERBOSE
      	swmerr ("it's a xa_windowHint");      
#endif//VERBOSE	
      } 
#ifdef CHECKCLIENTS      
      else
      { if (client->hints!=getValuedHint(client->window,xa_windowHints))
        { 
#ifdef ERRORMSG	
	  if (client->name)
	    swmerr ("client hints are wrong in propertyChangeHandler, %s",
	        client->name);
//#ifdef DEBUGPARANOID
//          exit(-1);
//#endif//DEBUGPARANOID	      
#endif//ERRORMSG	      
	  client->hints=getValuedHint(client->window,xa_windowHints);
	}//if hints
      }//else if event->atom	
#endif//CHECKCLIENTS      
#endif//GNOME

#ifdef NAMEHANDLING
      if (event->atom == XA_WM_NAME)
      {
#ifdef VERBOSE
        swmerr ("Xfreeing client->name in propertyChangeHandler");
#endif//VERBOSE
        if (client->name) XFree(client->name);


//was wrapped in ignoreXerrors
	status=XFetchName(display, client->window, &client->name);
	if (!status) 
	{
#ifdef ERRORMSG	
	  swmerr ("ERROR FETCHING NAME");
#endif//ERRORMSG	  
	  client->name=NULL;
	}  

	redrawParent(client);
      }//XA_WM_NAME
#endif//NAMEHANDLING
      if (event->atom == XA_WM_NORMAL_HINTS)
      { long dummy;
        XGetWMNormalHints(display, client->window, client->size, &dummy);

#ifdef MAXIMIZE
        client->maximized=false;
#endif//MAXIMIZE

        updateClient(client,UPDATESIZE);
#ifdef VERBOSE
        swmerr ("got XA_WM_NORMAL_HINTS"); 
#endif//VERBOSE
      }//XA_WM_NORMAL_HINTS 

#ifdef LAYERS	
      if (event->atom == xa_windowLayer)
      {
        oldLayer=client->layer;
        client->layer=getValuedHint(client->window,xa_windowLayer);

#ifdef VERBOSE
        swmerr ("adjusting Layer to: %d",client->layer); 
#endif//VERBOSE
// **************************************************************************
// FOR TESTING
// new:
        adjustLayer(client,(client->layer>oldLayer));

// old one did not work properly:
//        if ( client->layer > oldLayer) raiseClient(client);
//        else 			       lowerClient(client);
// **************************************************************************
      }// end if xa_windowLayer
#endif		

#ifdef GNOME	
      if (event->atom == xa_wm_desktop)
      {
#ifdef VERBOSE
        swmerr ("adjusting desktop");
#endif//VERBOSE
        moveClientToDesktop(client,getValuedHint(client->window,xa_wm_desktop));
      }
#endif		
  }// end if client 

#ifdef TRACE
  trace (-1,"propertyChangeHandler"); 
#endif
}//propertyChangeHandler 

// **************************************************************************

//  enterEventHandler
//! handles window enter events
/*!
\verbatim   
   v1.3.0-pre1: * initial release
                * should other enter events be ignored?
   v1.3.0-pre3: * for TESTING: in gnome mode clients get no focus if they don't 
                  want it
   v1.3.2:      * minimal mode now checking for childWindow directly				
   v1.3.3:	* now using caches version of hint
   v1.3.4:	* returned to "focus window" since not all appications like it
   v1.3.5:      * now testing both methods with FASTFOCUS
   v1.3.6:      * added SHOWFOCUS via focusClient to visualize focus behaviour
                * NOMOUSEFOCUS prevent Focus-Follows-Mouse
   v1.3.6:      * now using unified focusClient
    
\endverbatim   
\todo FOR TESTING: is event->window or client->window better??
*/
void enterEventHandler(XCrossingEvent *event)
{ 
#ifndef FASTFOCUS
  Client *client;
#endif//FASTFOCUS
  
#ifdef TRACE
  trace (1,"enterEventHandler"); 
#endif//TRACE

  // we may only handle the last event  
  while(XCheckTypedEvent(display, EnterNotify, (XEvent *)event));

#ifndef NOMOUSEFOCUS

  // childWindow is not enough if you want to select input boxes...
#ifndef FASTFOCUS
  client = getClient(event->window, anyWindow);
  
  if (client) 
  { 
#ifdef VERBOSE
    if (client->name) swmerr ("window: %s",client->name);
    swmerr ("clients window: %x",client->window);
    swmerr ("events  window: %x",event->window);
#endif//VERBOSE

#ifdef LOWCOLOR  
    XInstallColormap(display, client->cmap);
#endif//LOWCOLOR

#ifdef GNOME
    if(client->hints & WIN_HINTS_SKIP_FOCUS) 
    { 
#ifdef VERBOSE
      swmerr ("client wants no input focus\n");
#endif//VERBOSE
#ifdef CHECKCLIENTS      
      if (!(getValuedHint(client->window,xa_windowHints) &
          WIN_HINTS_SKIP_FOCUS))
	{ if (client->name)
	    swmerr ("window hints are not set correctly on client: %s",
	        client->name);  
		
	  client->hints=getValuedHint(client->window,xa_windowHints);
	}
#endif//CHECKCLIENTS
#ifdef TRACE
      trace (-1,"enterEventHandler"); 
#endif//TRACE
      return; 
    }  
#endif//GNOME
    focusClient(client);
  }//if client 
  else
  { 
#ifdef SWMPANEL  
    if (event->window!=sWMWindow)
    {
#endif//SWMPANEL
#ifdef ERRORMSG
    swmerr ("giving input focus to unknown window");
#endif//ERRORMSG
#ifdef SWMPANEL  
    }
#ifdef VERBOSE
    else swmerr ("giving input focus to SWMPANEL");
#endif//VERBOSE
#endif//SWMPANEL    
#endif//!FASTFOCUS

    // Now focusing unknown window
#ifdef MINIMAL
    XSetInputFocus(display, event->window, RevertToPointerRoot, CurrentTime);
#else//MINIMAL
    ignoreXErrors(XSetInputFocus(display, event->window, RevertToPointerRoot, CurrentTime));
#endif//MINIMAL

#ifndef FASTFOCUS
  }// end else if client
#endif//!FASTFOCUS  
  
#endif//NOMOUSEFOCUS
  
#ifdef TRACE
  trace (-1,"enterEventHandler"); 
#endif//TRACE
}//enterEventHandler 

// ***************************************************************************

//  exposeEventHandler
//! redraws windows on expose event
/*! this redraw is simple: it will look for the last expose event only and 
    redraw the entire window
\verbatim   
   v1.3.0-pre1: initial release
   v1.3.0-pre3: changes to speed up exposure, getClient now only called for
                the last exposeEvent
   v1.3.0-pre4: added debug
   v1.3.5:      now returning before debug part if many exposes are pending
   v1.3.6:	added support for the sWM Panel window
\endverbatim   
*/
void exposeEventHandler(XExposeEvent *event)
{ Client *client;

#ifdef TRACE
  trace (1,"exposeEventHandler"); 
#endif

 if (event->count>0) 
 {
#ifdef TRACE
   trace (-1,"exposeEventHandler (still exposes pending)"); 
#endif
   return;
 }  
   
// DEBUG - PART 
#ifdef VERBOSE
  client = getClient(event->window, anyWindow);
  if (client) 
  { if (client->name) swmerr ("window: %s",client->name);
  }
#endif//VERBOSE 

// FUNCTION - PART 

#ifdef SWMPANEL
  if (event->window==sWMWindow) 
  { refreshPanel(); 
  }
  else
  {
#endif//SWMPANEL

    // only look for client and redraw if it is really the last expose-event
    client = getClient(event->window, parentWindow);

    if (client)
    {
#ifdef VERBOSE
      if (client->name) swmerr ("redrawing window: %s",client->name);
#endif//VERBOSE
      redrawParent(client);
    }// end if client
#ifdef SWMPANEL
  }
#endif//SWMPANEL  
#ifdef TRACE
  trace (-1,"exposeEventHandler"); 
#endif//TRACE
}//ExposeEventHandler 

// ***************************************************************************
#ifdef VERBOSE
//  reparentEventHandler
//! currently does nothing
void reparentEventHandler(XReparentEvent *event)
{
#ifdef TRACE
  trace (0,"reparentEventHandler(*) (but doing nothing)"); 
#endif//TRACE

}//reparentEventHandler 
#endif//VERBOSE
// ***************************************************************************

//  destroyNotifyHandler
//! tries to remove the client that was destroyed
/*!
\verbatim
   v1.3.0-pre1: initial release
   v1.3.0-pre4: added debug
   v1.3.0-pre5: now using removeParent
\endverbatim   
*/
void destroyNotifyHandler (XDestroyWindowEvent *event)
{ Client *client;
   
#ifdef TRACE
  trace(1,"destroyNotifyHandler"); 
#endif//TRACE

// DEBUG - PART 
#ifdef VERBOSE
  client = getClient(event->window, anyWindow);
  if (client) 
  { if (client->name) swmerr ("window: %s",client->name);
  }
#endif//VERBOSE 
 
// FUNCTION - PART  

  // only destroy a client if really the child window is destroyed
  client=getClient(event->window, childWindow); 
  if(client)
  { 
    removeParent (client, false); // Don't remap again
  }
#ifdef ERRORMSG
  else
  { swmerr ("ERROR: no client, seems to be destroyed already");
  }
#endif//ERRORMSG

#ifdef TRACE
  trace(-1,"destroyNotifyHandler"); 
#endif//TRACE

}//destroyNotifyHandler

// ***************************************************************************

//  configureRequestHandler
//! handles configure requests
/*!
\verbatim   
   v1.3.0-pre1: intial release
   v1.3.1-pre2: fixed silly else,else bug that prevents clients to move
                vertically
   v1.3.1-pre3: fixed double check for window
   v1.3.1:      fixed layer management, testing sibling if no layers
\endverbatim   
\todo event->above seems to hold trash - or do I just not understand it?
*/
void configureRequestHandler(XConfigureRequestEvent *event)
{ Client *client;
#ifdef LAYERS
  Client *sibling;
#endif
  XWindowChanges wc;

#ifdef TRACE
  trace (1,"configureRequestHandler"); 
#endif

  client = getClient(event->window, childWindow);

  if (client) 
  { 
#ifdef VERBOSE
    if (client->name) swmerr ("window: %s",client->name);
#endif//VERBOSE
  
#ifdef MWM
    if (client->frameState==FRAMESHOWN)
#endif // MWM

      changeGravity(client,-1);

    if (event->value_mask & CWX) 
    { client->x = event->x;
#ifdef RESTRICTAUTORESIZE
      if (client->x < 0 ) client->x=0;
      if (client->x>xmax) client->x=xmax-1;
#endif	
    }

    if (event->value_mask & CWY) 
    { client->y = event->y;
#ifdef RESTRICTAUTORESIZE
      if (client->y < 0) client->y=0;
      if (client->y>ymax) client->y=ymax-1;    
#endif
    }

    if (event->value_mask & CWWidth) 
    { client->width = event->width;
#ifdef RESTRICTAUTORESIZE
      if (client->width+client->x>xmax && (client->x<xmax)) 
        client->width=xmax-client->x;
#endif	  
    }
    
    if (event->value_mask & CWHeight) 
    { client->height = event->height;
#ifdef RESTRICTAUTORESIZE
      if (client->height+client->y>ymax && (client->y<ymax)) 
        client->height=ymax-client->y;
#endif	  
    }
    
#ifdef MWM
    if (client->frameState==FRAMESHOWN)
#endif // MWM
      changeGravity(client,1);

#ifdef ALWAYSCENTER
    client->x=xmax/2-client->width/2;
    client->y=ymax/2-client->height/2;
#endif

    if (event->value_mask & (CWX|CWY|CWWidth|CWHeight))
      updateClient(client,UPDATESIZE);

    if (event->value_mask & CWStackMode)
    { 
#ifdef VERBOSE
      swmerr ("adjusting stacking order");
#endif//VERBOSE  
      wc.stack_mode = event->detail;
#ifdef LAYERS      
      sibling       = getBoundingClient(client->layer,(wc.stack_mode==Above)); 
      if (sibling) wc.sibling = sibling->parent;
//    FOR TESTING: event->above seems to hold trash
//      else         
#else
      wc.sibling = event->above;
#endif
      if (wc.sibling)
       XConfigureWindow(display,client->parent, CWStackMode|CWSibling,&wc);
      else  
        XConfigureWindow(display,client->parent,CWStackMode,&wc);
    }// end if raised/lowered
  }// end if (client) 
  else
  { 
#ifdef DEBUG
    swmerr ("configuring unknown client / other part of window");
    client = getClient(event->window, anyWindow);
    if (client) 
    { 
      if (client->name) 
        swmerr ("was a client (%s)",client->name);
      else 
        swmerr ("was a client without name");
    }	
#endif//DEBUG

    wc.x 	= event->x;
    wc.y 	= event->y;
    wc.width 	= event->width;
    wc.height 	= event->height;
    wc.sibling 	= event->above;
    wc.stack_mode = event->detail;
#ifdef MINIMAL
    XConfigureWindow(display,event->window,event->value_mask,&wc);
#else
    ignoreXErrors(XConfigureWindow(display,event->window,event->value_mask,&wc));
#endif //MINIMAL
  }
#ifdef TRACE
  trace (-1,"configureRequestHandler"); 
#endif  
}// end configureRequestHandler 

// ***************************************************************************

#ifdef KEYS
//  keyEventHandler
//! manages all incoming key events for sWM,
/*! keys are usually grabbed to the rootWindow so we have to be carefull,
    which window we modify - it may be on another desktop  
\verbatim
   v1.3.0-pre1: initial release
   v1.3.0-pre2: window close: moved next-raised stuff to sendExitClient
   v1.3.0-pre3: added exact keys
   v1.3.0-pre4: changed from focusedClient to getFocusedClient
   v1.3.2:      now using keybindings
   v1.3.3:      updated to expose after layer switching,
		now using "subwindow" to determine window, if it is not
		valid, using focusedWindow
   v1.3.5:	now checking if client is really on current desktop,
   		if no client is focused, window dependent checks will not
		be run
   v1.3.6-pre3: setting input focus to cycled client - if desired
   v1.3.6:	added maximize using Alt-F6 per default,
		when no window is focused, the handler exits after window
		independent checks

\endverbatim		
*/
void keyEventHandler (XKeyEvent *event)
{ Client *client;
  int 	 handled=false;
  int    eventstate = event->state;

#ifdef TRACE
  trace(1,"keyEventHandler");
#endif//TRACE

  if (!event) 
  {
#ifdef ERRORMSG
    swmerr ("error, no event given in keyEventHandler");
#endif//ERRORMSG
  
#ifdef TRACE
    trace(-1,"keyEventHandler");
#endif//TRACE  

#ifdef DEBUGPARANOID
    exit(-1);
#endif//DEBUGPARANOID
    return;
  }//!event
  
  if (event->window==rootWindow) 
  {

#ifndef EXACTKEY  
#ifdef VERBOSE
    swmerr ("filtered keys");
#endif //VERBOSE
    eventstate &= (Mod1Mask|ControlMask);
#endif//EXACTKEY

// ****** DEBUG - PART ************
#ifdef DEBUG
    // the check HERE is only nessesary for debugging
    client=getClient(event->subwindow,anyWindow);
    if (!client) client=getFocusedClient();

    if (event->keycode==XKeysymToKeycode(display,XStringToKeysym(DEBUGKEY)) 
        && eventstate == Mod1Mask)		
    { Window focusedWindow;
      int dummy;
      dumpClients();
#ifndef DEBUGPARANOID
      nodebug++;
#endif//DEBUGPARANOID

      client=getFocusedClient();
      XGetInputFocus(display,&focusedWindow, &dummy);
#ifndef DEBUGPARANOID
      nodebug--;
#endif //DEBUGPARANOID     
      if (client->name)
        swmerr ("focused: %s (clientsWin: %#x, reallyFocusedWin: %#x)",
	    client->name,client->window,focusedWindow);
    }// end Alt-DEBUGKEY

    client=NULL;
#endif//DEBUG       

// ****** FUNCTION - PART ************
#ifdef VERBOSE
    swmerr ("beginning function part");
#endif //VERBOSE

// ************************************************************************
// Window independent checks
// ************************************************************************

// ************* Alt-CYCLEKEY switches to the next window ***************
#ifndef NORAISE
    if (event->keycode == XKeysymToKeycode(display, XStringToKeysym(CYCLEKEY)) 
        && eventstate == Mod1Mask) 
    { 
#ifdef GNOME
      raiseNextClient();
#ifdef VERBOSE
      swmerr ("finished raiseNextClient");
#endif//VERBOSE

#ifdef CYCLEFOCUS
      if (raisedClient) 
      { focusClient(raisedClient);
      }
#endif//CYCLEFOCUS

#ifdef VERBOSE
      swmerr ("finished focusing");
      if (raisedClient) 
      { if (raisedClient->name) swmerr ("changed to: %s",raisedClient->name);
        client=getClient(raisedClient->window,childWindow);
        if (client && client->name) 
	  swmerr ("changed to (per getClient): %s",client->name);      
      } else { swmerr ("changed to no other client"); }
#endif//VERBOSE

#else//GNOME 
      XCirculateSubwindowsUp(display,rootWindow);
#endif//GNOME 
      handled=true;
    } //if Alt-CYCLEKEY 
#endif//NORAISE


// ************ Ctrl-CYCLEKEY switches to the next desktop *****************
#ifdef GNOME
    if (event->keycode == XKeysymToKeycode(display, XStringToKeysym(CYCLEKEY)) 
        && eventstate == ControlMask) 
    { 
#ifdef VERBOSE
      swmerr ("changing from desktop: %i\n",currentDesktop);
#endif//VERBOSE
      if (currentDesktop>NUMBEROFDESKTOPS-2) changeDesktop(0);
      else changeDesktop (currentDesktop+1);
#ifdef VERBOSE
      swmerr ("changed to desktop: %i\n",currentDesktop);
#endif//VERBOSE
      handled=true;
    }// end if Ctrl-CYCLEKEY
#endif//GNOME

// ********************** Ctrl-Alt-EXITKEY exits sWM ************************
#ifdef EXITKEY
    if (event->keycode == XKeysymToKeycode(display, XStringToKeysym(SWMEXITKEY)) 
        && eventstate == Mod1Mask) 
    { 
#ifdef VERBOSE
      swmerr ("got exitkey, exiting sWM\n");
#endif//VERBOSE
      sWMQuitNicely();
    }// end if Ctrl-Alt-EXITKEY
#endif//EXITKEY


#ifdef VERBOSE
      swmerr ("finished window independent checks");
#endif//VERBOSE

// ***************************************************
// Window dependent checks

// this we already did if we are in debug mode

    if (!handled)
    { client=getClient(event->subwindow,anyWindow);
      if (!client) client=getFocusedClient();

      // if really no client is focused, leave now
      if (!client) return; // sorry, we have tried...

#ifdef VERBOSE
      if (client->name) swmerr ("window: %s",client->name);
#endif//VERBOSE

// if the window is not on the current desktop, we can quit now
#ifdef GNOME
      if (client->desktop!=currentDesktop) 
      {
#ifdef ERRORMSG
        swmerr ("ERROR: client on other desktop got keyboard focus (FIXME)");
#endif//ERRORMSG

#ifdef TRACE
        trace(-1,"keyEventHandler");
#endif//TRACE
        return;
      }
#endif//GNOME    

// ********************************************************************
// OK, one of our windows on the current desktop got a keyboard command
// ********************************************************************

// ******* RAISE layer *************
#ifdef LAYERS
      if (event->keycode==XKeysymToKeycode(display,XStringToKeysym(LAYERUP))
          && eventstate == Mod1Mask)
      { // now using subwindow instead of getFocusedClient
        // client=getFocusedClient();
      
        // layer=getValuedHint(client->window, xa_windowLayer);
        setLayer (client, client->layer+1);
        redrawParent(client);
      }// end Alt-LAYERUP 


// ******* LOWER layer *************
      if (event->keycode==XKeysymToKeycode(display,XStringToKeysym(LAYERDOWN))
          && eventstate == Mod1Mask)
      { // now using subwindows instead of getFocusedClient
        // client=getFocusedClient();
        // layer=getValuedHint(client->window, xa_windowLayer);
        setLayer (client, client->layer-1);
        redrawParent(client);
      }//Alt-LAYERDOWN     
#endif//LAYERS
		
										
// ********* FRAME/UNFRAME window ************
#ifdef MWM    
      if (event->keycode==XKeysymToKeycode(display,XStringToKeysym(FRAMEKEY)) 
          && eventstate == Mod1Mask)		
      { // toggleFrameHidden(getFocusedClient());
        toggleFrameHidden(client);
      }// end Alt-FRAMEKEY
#endif//MWM
    	    
// ********* close window ************
      if (event->keycode==XKeysymToKeycode(display,XStringToKeysym(CLOSEKEY)) 
          && eventstate == Mod1Mask)		
      { 
#ifdef VERBOSE
        if (client->name)
	  swmerr ("got Alt+",CLOSEKEY,", closing window: %s",client->name); 
	if (getFocusedClient()->name)
          swmerr ("focused is window: %s",getFocusedClient()->name); 
#endif//VERBOSE
        sendExitClient(client);
      }//Alt-CLOSEKEY          

#ifdef KEYRAISE
// ******** RAISING window ***********
      if (event->keycode==XKeysymToKeycode(display,XStringToKeysym(RAISEKEY)) 
          && eventstate == Mod1Mask)		
      { 
#ifdef VERBOSE
        if (client->name)
	  swmerr ("got Alt+",RAISEKEY,", raising window: %s",client->name); 
	  
	if (getFocusedClient()->name)
          swmerr ("focused is window: %s",getFocusedClient()->name); 
#endif//VERBOSE

#ifdef NOMOUSEFOCUS
         focusClient(client); 
#endif//NOMOUSEFOCUS
         raiseClient(client);
      }// end if Alt-RAISEKEY 
#endif//KEYRAISE


#ifdef KEYMAXIMIZE
// ******** MAXIMIZE window ***********
      if (event->keycode==XKeysymToKeycode(display,XStringToKeysym(MAXIMIZEKEY)) 
          && eventstate == Mod1Mask)
      { 
#ifdef VERBOSE
        if (client->name) 
	  swmerr ("got Alt+",MAXIMIZEKEY,", resizing to full screen: %s",
	       client->name); 
        if (getFocusedClient()->name)
	  swmerr ("focused is window: %s",getFocusedClient()->name); 
#endif//VERBOSE

	 if (client->maximized)
	    unmaximizeClient(client); 
	 else
            maximizeClient(client); 
      }//Alt-MAXIMIZEKEY 
#endif//KEYMAXIMIZE

    }//!handled

  }// end if rootWindow 
  else
  { 
#ifdef ERRORMSG
      swmerr ("*** got a keyboard-event?!? sWM should not receive it that way! ***");  
#endif//ERRORMSG
  }// end if event->window

#ifdef TRACE
  trace(-1,"keyEventHandler");
#endif // TRACE
}//keyEventHandler
#endif//KEYS

// *************************************************************************

//  buttonPressedEventHandler
//! handles button presses
/*!
\verbatim
   v1.3.0-pre1: initial release
   v1.3.0-pre3: fixed a bug preventing to change to the first desktop 
   v1.3.0-pre4: updated some "click to raise" functions (beginning)
   v1.3.1:      eliminated getClickedItem in MINIMAL mode
   v1.3.2:      eliminated backmove in MINIMAL mode
   v1.3.6:      changing focus if titlebar is clicked (!NOMOUSEFOCUS)
\endverbatim   
*/
void buttonPressedEventHandler(XButtonEvent *event)
{ Client *client; 
  int clickedItem;
#ifdef CLICKDESK
  char stringDummy[100];
#endif//CLICKDESK

#ifdef TRACE
  trace(1,"buttonPressedEventHandler");
#endif//TRACE

 // if X has been busy, just take the last action to avoid unexpected behaviour 


#ifndef MINIMAL
 while (XCheckTypedEvent(display,ButtonPress,(XEvent *) event));

 if (event->window == rootWindow) 
 {
#ifndef CLICKDESK
   switch (event->button) 
   { case Button1: system(CONFDIR "/.sWM/button1" "&"); break;
     case Button2: system(CONFDIR "/.sWM/button2" "&"); break;
     case Button3: system(CONFDIR "/.sWM/button3" "&"); break; 
   }
#else//CLICKDESK
   sprintf (stringDummy,"%s/.sWM/button%d %d %d &",
            CONFDIR,event->button,
	    1+(event->x/CLICKDESKX),
	    1+(event->y/CLICKDESKY));
   system  (stringDummy);
#endif//CLICKDESK   
 } 
 else 
#endif//MINIMAL 
 { client = getClient(event->window, anyWindow);
   if (client) 
   {
#ifdef VERBOSE
     if (client->name) swmerr ("window: %s",client->name);
#endif//VERBOSE

#ifdef RAISEONCLICK
     if (event->subwindow == client->window)
     { 
#ifndef NORAISE
       if (client!=raisedClient) 
#endif//NORAISE       
       { raiseClient (client);  
         XUngrabPointer (display, CurrentTime);
         XSendEvent(display, sWMWindow, false, SubstructureNotifyMask, 
                    (XEvent *) event);
       }
       XAllowEvents (display, ReplayPointer, CurrentTime);
       XAllowEvents (display, SyncPointer, CurrentTime);
     }
     else
     {
#endif//RAISEONCLICK 

// 0 should be normal titlebar, 1 closebox, 2 maxbox, 3 sticky, 4 windowbody
#ifdef PIXMAPS
     clickedItem=getClickedItem(client,event->window,event->y);
#else//PIXMAPS 

#ifdef MINIMAL
     clickedItem=((client->width-event->x)/TITLEBARHEIGHT)+1;
     if (clickedItem!=CLOSEBOX) clickedItem=TITLEBAR;
#else//MINIMAL
     clickedItem=getClickedItem(client,event->x,event->y);
#endif//MINIMAL
#endif//PIXMAPS 
	     
#ifdef VERBOSE
       swmerr ("clicked item: %#x",clickedItem);
#endif//VERBOSE
       motionStartingX=event->x_root-client->x;
       motionStartingY=event->y_root-client->y;     

       switch (event->button) 
       {
         case Button1:
#ifdef GNOME
             if (clickedItem==STICKYBOX)
             { toggleClientSticky(client);
             }
#endif//GNOME     

#ifndef ALWAYSMAX
#ifdef MAXIMIZE
             if (clickedItem==MAXBOX)
             { maximizeClient(client);
             }
#endif//MAXIMIZE 
#endif//!ALWAYSMAX
             if (clickedItem==CLOSEBOX)
	     { sendExitClient(client);
	     }// end if CLOSEBOX
	     
             if (clickedItem==TITLEBAR)
             { 
#ifdef NOMOUSEFOCUS
               focusClient(client);
#endif//NOMOUSEFOCUS
	     
	       //Beginning window movement, raising it
	       raiseClient(client);
#ifndef ALWAYSCENTER
#ifndef ALWAYSMAX	       
	       moveClient(client);
#endif//!ALWAYSMAX
#endif//ALWAYSCENTER	       
             }
#ifndef MINIMAL	  
#ifndef ALWAYSMAX   
             if (clickedItem==WINDOWBODY)
             { //Beginning window movement, raising it
#ifndef ALWAYSCENTER
	       moveClient(client);
#endif//!ALWAYSCENTER	       
             }
#endif//!ALWAYSMAX	     
#endif//MINIMAL	     
             break;

#ifndef MINIMAL
        case Button2: 
             if (clickedItem==CLOSEBOX) 
	     { lowerClient(client);
	     }
#ifdef SHADE	   
             if (clickedItem==TITLEBAR) 
	     { 
	       toggleClientShaded(client); 
	     } 
#endif//SHADE   
	     if (clickedItem==WINDOWBODY)
	     { raiseClient (client);
#ifndef ALWAYSMAX   
#ifndef ALWAYSCENTER
	       moveClient  (client);
#endif//!ALWAYSCENTER	       
#endif//!ALWAYSMAX	       
	     } 
#ifdef GNOME
             if (clickedItem==STICKYBOX)
             { if (currentDesktop==0)
	        moveClientToDesktop(client,NUMBEROFDESKTOPS-1);
	       else 
	        moveClientToDesktop(client,currentDesktop-1);
             }
#endif//GNOME    	     	     
	     break;
#endif//MINIMAL 
        case Button3: 
#ifndef ALWAYSMAX   
             if (clickedItem==CLOSEBOX) 
	     { //Beginning window resize
#ifdef NORESIZEWARP
	       handledClient_old_y=client->y+client->height;
#endif//NORESIZEWARP	     
	       resizeClient (client);	   
	     }
#ifdef MAXIMIZE		   	     
	     if (clickedItem==MAXBOX) 
	     { 
	       unmaximizeClient(client);
	     } 
#endif//MAXIMIZE	     

#ifndef MINIMAL
#ifndef ALWAYSCENTER
	     if (clickedItem==TITLEBAR)
	     { // Beginning window movement in background
               moveClient (client);
	     } 
#endif//!ALWAYSCENTER	     
	     if (clickedItem==WINDOWBODY)
	     { // Beginning window resizement im hintergrund 
#ifdef NORESIZEWARP
	       handledClient_old_y=false;
#endif//NORESIZEWARP
               resizeClient (client);
	     } 
#endif//MINIMAL
#endif//!ALWAYSMAX

#ifdef GNOME
             if (clickedItem==STICKYBOX)
             { if (currentDesktop<NUMBEROFDESKTOPS-1)
	        moveClientToDesktop(client,currentDesktop+1);
	       else 
	        moveClientToDesktop(client,0);
             }
#endif//GNOME    
	     break;
        }//switch button 
#ifdef RAISEONCLICK       
      }//if not client window 
#endif//RAISEONCLICK
    }//if client 
#ifdef SWMPANEL
    else
    { if (event->window==sWMWindow && event->button==Button1)
      { panelClick(event->x,event->y);
      }
#ifdef SWMPANELHIDE      
      if (event->window==sWMWindow && event->button==Button2)
      { togglePanelHidden();
      } 
#endif//SWMPANELHIDE      
    }
#endif//SWMPANEL    
  }//else if rootwin 

#ifdef TRACE
  trace(-1,"buttonPressEventHandler");
#endif//TRACE
}//buttonPressEventHandler 


// -mouseMotionEventHandler------------------------------------------------
//! handles mouse movement while moveing or resizing clients
/*!
\verbatim
   v1.3.0-pre1: initial release
   v1.3.0-pre2: window move/resize now uses updateClient directly
   v1.3.0-pre3: updated to support FAST again
   v1.3.1-pre1: added NORESIZEWARP
\endverbatim
*/
void mouseMotionEventHandler(XMotionEvent *event)
{
#ifdef TRACE
  trace(1,"mouseMotionEventHandler");
#endif//TRACE

#ifndef MINIMAL
  while (XCheckTypedEvent (display, MotionNotify, (XEvent *) event));
#endif//MINIMAL

  if (handledClient) 
  { if (pointerMode==POINTERMOVEMODE)
    { 
#ifdef FAST
      drawOutline(handledClient);
#endif//FAST      

      handledClient->x=event->x_root-motionStartingX;
      handledClient->y=event->y_root-motionStartingY;

#ifndef MINIMAL
#ifndef FAST 
      updateClient (handledClient,UPDATESIZE);	
#else//FAST
      drawOutline(handledClient);
#endif//FAST	
#endif//MINIMAL 
    }// end if POINTERMOVEMODE
    if (pointerMode==POINTERRESIZEMODE)
    { 
#ifdef FAST
      drawOutline(handledClient);
#endif//FAST	
      handledClient->width=event->x_root-handledClient->x;
      
      if (handledClient->width<MINWINDOWWIDTH) handledClient->width=MINWINDOWWIDTH;
      
      handledClient->height=event->y_root-handledClient->y;
      
      if (handledClient->height<0) handledClient->height=0;

#ifndef MINIMAL
#ifndef FAST
      updateClient (handledClient,UPDATESIZE); 
#else//FAST
      drawOutline (handledClient);
#endif//FAST	
#endif//MINIMAL 
    }//if POINTERRESIZEMODE
#ifdef NORESIZEWARP    
    if (pointerMode==POINTERBOXRESIZEMODE)
    { 
#ifdef FAST
      drawOutline(handledClient);
#endif	
      handledClient->width = event->x_root-handledClient->x;
      if (handledClient->width<MINWINDOWWIDTH) handledClient->width=MINWINDOWWIDTH;
      
      handledClient->y=event->y_root+TITLEBARHEIGHT;
      handledClient->height=handledClient_old_y-event->y_root-TITLEBARHEIGHT;
      if (handledClient->height<0) handledClient->height=0;

#ifndef MINIMAL
#ifndef FAST
      updateClient (handledClient,UPDATESIZE); 
#else
      drawOutline (handledClient);
#endif	
#endif // MINIMAL
    }
#endif // NORESIZEWARP    
  }// end if andledClient 
#ifdef TRACE
  trace(-1,"mouseMotionEventHandler");
#endif

}//mouseMotionEventHandler 


// -buttonReleaseEventHandler----------------------------------------------
//! handles button releases and finishes running actions
/*!
\verbatim
   v1.3.0-pre1: initial release
   v1.3.0-pre3: updated to support FAST again, skiping closed window
\endverbatim   
*/
void buttonReleaseEventHandler(XButtonEvent *event)
{
#ifdef TRACE
  trace(1,"buttonReleaseEventHandler");
#endif

   if (handledClient) 
   {
#ifdef VERBOSE
     if (handledClient->name) swmerr ("window: %s",handledClient->name);
#endif//VERBOSE

     if (pointerMode!=POINTERNORMALMODE)
     { pointerMode=POINTERNORMALMODE;
       XUngrabPointer(display,CurrentTime);
#ifdef FAST
       // clear window border again
#ifndef MINIMAL
       drawOutline(handledClient);
       // We have to send an expose event - otherwise the window will be distorted
#endif//MINIMAL
#endif//FAST      
       updateClient(handledClient,UPDATESIZE);
       handledClient=NULL;
     }//if pointernormalmode   
   }//if handledClient 

#ifdef TRACE
  trace(-1,"buttonReleaseEventHandler");
#endif//TRACE

}//buttonReleaseEventHandler 

// *************************************************************************

//  sWMEventLoop
//! is the main part of sWM, receiving all event
/*!
\verbatim   
   v1.3.0-pre1: initial release
   v1.3.0-pre4: showEvents is now VERBOSE
   v1.3.2:      now skipping clientMessageHandler
   v1.3.4:	only running reparentEventHandler in VERBOSE mode
\endverbatim
*/

void sWMEventLoop()
{ XEvent event;
#ifdef TRACE
  trace(1,"sWMEventLoop");
#endif//TRACE

  while (!exitSWM) 
  {   if (clientToRemove) removeParent(clientToRemove,false);
      clientToRemove=NULL;
      
      XNextEvent(display, &event); 

#ifdef VERBOSE
      showEvents(event);
#endif//VERBOSE
      switch (event.type) 
      { case ButtonPress  : buttonPressedEventHandler  	(&event.xbutton); 
             break;
        case ButtonRelease: buttonReleaseEventHandler	(&event.xbutton); 
             break;
        case MotionNotify : mouseMotionEventHandler    	(&event.xmotion);    
             break; 
        case EnterNotify  : enterEventHandler	     	(&event.xcrossing); 
             break;
        case ConfigureRequest:configureRequestHandler	(&event.xconfigurerequest); 
             break;
        case MapRequest	  : mapRequestHandler	   	(&event.xmaprequest);  
             break;
        case UnmapNotify  : unmapEventHandler	   	(&event.xunmap); 
             break;
#ifndef MINIMAL	     
        case ClientMessage: clientMessageHandler	(&event.xclient); 
             break;
#endif//MINIMAL
        case PropertyNotify:propertyChangeHandler    	(&event.xproperty); 
             break;
        case Expose       : exposeEventHandler 	   	(&event.xexpose); 
             break;
#ifdef VERBOSE
        case ReparentNotify: reparentEventHandler	(&event.xreparent); 
             break;
#endif//VERBOSE
        case DestroyNotify: destroyNotifyHandler 	(&event.xdestroywindow); 
	     break;
#ifdef LOWCOLOR      
        case ColormapNotify: 
	                    colormapChangeHandler   	(&event.xcolormap); 
             break;
#endif//LOWCOLOR
#ifdef KEYS
	case KeyPress  	  : 
	                    keyEventHandler  		(&event.xkey); 
             break;
#endif//KEYS
      }// end switch 
  }// end while not exitSWM 

#ifdef TRACE
  trace(-1,"sWMEventLoop");
#endif//TRACE
}//sWMEventLoop 

// ***************************************************************************
