/*!
\file tools.c
\author Robert Sperling <sperling@small-window-manager.de>
\version 1.3.6
\brief essential things for handling internal data structures

\verbatim
 ***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************
\endverbatim 
*/

// -getClient--------------------------------------------------------------- 
//! finds the client managing a given window
/*! 
\returns the client which belongs to the given window
\param givenWin 	the window affected
\param whichWindow	one of anyWindow, childWindow, parentWindow

\verbatim
   v1.3.0-pre1: initial release
   v1.3.0-pre2: changed checking order, added anyWindow
   v1.3.0-pre3: added pixmap support
   v1.3.0-pre4: bugs are always where you can't find them. getClient has
                been too good in DEBUGPARANOID, some clients should not
		be found! (unmapping of buttons in toggleFrameHidden)
   v1.3.2:      made it less verbose in ordinary debug mode
   v1.3.6:      now getClients checks for pixmap-boxes only in CHECKCLIENTS
                mode, giving a speedup if many clients are handled,
		givenWin is only checked in CHECKCLIENTS mode, since all other
		functions check it before
\endverbatim
*/
Client *getClient(Window givenWin, int whichWindow)
{ Client *client;

#ifdef TRACE
  trace (0,"getClient(*)"); 
#endif//TRACE

// invalid windows are rather uncommon so we don't check per default
#ifndef CHECKCLIENTS
  if (givenWin)
  { 
#endif//CHECKCLIENTS
    if (givenWin != rootWindow)
    { if (whichWindow==parentWindow)
      { for (client = headClient; client; client = client->next)
        { if (givenWin == client->parent) return client;
#ifdef PIXMAPS
#ifdef CHECKCLIENTS
          if (givenWin == client->closebox) 
	  { 
	    swmerr ("getClient found CLOSEBOX");
	    return client;
	  }
#ifdef MAXIMIZE
	  if (givenWin == client->maxbox) 
	  { swmerr ("getClient found MAXBOX");
	    return client;
	  }  
#endif//MAXIMIZE

#ifdef GNOME
	  if (givenWin == client->stickybox) 
	  { swmerr ("getClient found CLOSEBOX");
	    return client;
	  }  
#endif//GNOME
#endif//CHECKCLIENTS	  
	  
#endif//PIXMAPS
	}//for 
      }//if parentwindow
      else if (whichWindow==childWindow)
      { for (client = headClient; client; client = client->next)
          if (givenWin == client->window) return client;
      }
      else if (whichWindow==anyWindow)
      { for (client = headClient; client; client = client->next)
        { if (givenWin == client->window) return client;
          if (givenWin == client->parent) return client;
#ifdef PIXMAPS
          if (givenWin == client->closebox) return client;
#ifdef MAXIMIZE
	  if (givenWin == client->maxbox) return client;
#endif//MAXIMIZE	  

#ifdef GNOME
	  if (givenWin == client->stickybox) return client;
#endif//GNOME

#endif//PIXMAPS 	  
        }//for 
      }//if anywindow 
    }//if not root window 
#ifndef CHECKCLIENTS
  }//if window given
#ifdef ERRORMSG
  else
  { swmerr ("ERROR: getClient: no valid window given");
#ifdef DEBUGPARANOID
    exit(-1);
#endif//DEBUGPARANOID  
  }//if no window given
#endif//ERRORMSG  
#endif//CHECKCLIENTS

#ifdef VERBOSE
  swmerr ("have not found client");   
#endif//VERBOSE  

  return NULL;
}//getClient


// -getFocusedClient--------------------------------------------------
//! gets the currently focused client
/*!
\returns the client currently focused 

\verbatim
History
-------

   v1.3.0-pre1: initial release 
   v1.3.0-pre2: no change
   v1.3.0-pre3: changed to anyWindow
   v1.3.0-pre4: added test for correct focusedClient
   v1.3.3:	client was not pre-initialized  
   v1.3.6:      now focus check is only in CHECKCLIENTS
\endverbatim   
*/ 
#ifdef KEYS
Client *getFocusedClient() 
{ 
  Client *client = NULL; 
  Window  focusedWindow = 0; 
  int     dummyInt; 
 
#ifdef TRACE 
  trace (0,"getFocusedClient(*)"); 
#endif//TRACE 
 
  XGetInputFocus(display, &focusedWindow, &dummyInt); 
  
  // this check is needed, getClient won't check for valid windows
  if (focusedWindow)
  { 
     client = getClient (focusedWindow, anyWindow); 
    // testing 20040229
    //   client = getClient (focusedWindow, childWindow); 

    // It seems a client is really focused
    if (client)
    {
#ifdef NAMEHANDLING    
#ifdef VERBOSE
      if (client->name) swmerr ("focusedClient: %s",client->name);
      else swmerr ("unnamed focusedClient");
#endif//VERBOSE    
#endif//NAMEHANDLING
     
#ifdef CHECKCLIENTS     
      if (focusedWindow!=client->window)
      {
#ifdef ERRORMSG
        swmerr ("error: client window is not correctly focused");
#endif//ERRORMSG    
        // sometimes it happens that a wrong part of the client is focused
	// it will now be corrected
        focusClient(client);
      }// end if client not correctly focused 
#endif//CHECKCLIENTS      
    }// end if client
    else
    {
#ifdef VERBOSE
      swmerr ("no client focused");
#endif//VERBOSE    
    }
  }// end if window is not valid
  else
  {
#ifdef VERBOSE
    swmerr ("got no focused window");
#endif//VERBOSE   
  }      
  return client;   
}//getFocusedClient  
#endif//KEYS 

// -getNextFocusClient-----------------------------------------------------
//! gets next focus-able client out of the list
/*! gets the next client out of the list of clients, if it does not want the
    focus the next will be choosen - until a client is found or the current
    is selected again

\returns the client focused after this functions 

\verbatim
History
-------

   v1.3.0-pre1: initial release
   v1.3.0-pre2: changed to avoid deadlocks
   v1.3.0-pre3: no change
   v1.3.0-pre4: eliminated focusedClient
   v1.3.1:	eliminated old ->destroyed
   v1.3.3:      oops, counter was uninitialized
   v1.3.6:      now using unified focusClient
\endverbatim   
*/ 
#ifdef KEYS
Client *getNextFocusClient()
{ Client *client;
  Client *oldFocused;  

#ifdef GNOME
  int     finished=false;
#endif//GNOME 
  
#ifdef TRACE 
  trace (1,"getNextFocusedClient"); 
#endif//TRACE 

  client              = getFocusedClient();  
  if (!client) client = headClient;
  oldFocused          = client;

#ifdef GNOME 
  while (!finished)
  { 
#endif//GNOME
 
    if (client->next) 
      client=client->next;
    else
      client=headClient;

#ifdef GNOME    
    //next client in GNOME mode is next on this desktop or the client itself
    if(!getValuedHint(client->window, xa_windowHints)&WIN_HINTS_SKIP_FOCUS) 
      if (client->desktop==currentDesktop) finished=true; 
     
    if (client==oldFocused) finished=true;
  }//while not finished 
#endif//GNOME  

  focusClient(client);
  
#ifdef TRACE 
  trace (-1,"getNextFocusedClient"); 
#endif//TRACE 

  return client;
}// end getNextFocusedClient 
#endif//KEYS 

// -focusClient------------------------------------------------------------
//! put keyboard focus to the given client
/*! focused given client and if SHOWFOCUS is enabled, draws a bigger border
    around the window
\verbatim
   v1.3.6-pre2: initial release
\endverbatim   
*/ 

void focusClient(Client *client)
{
#ifdef SHOWFOCUS
    if (lastFocusedClient && (lastFocusedClient!=client))   
      XSetWindowBorder(display, lastFocusedClient->parent, borderColor.pixel); 

    lastFocusedClient=client;
    XSetWindowBorder(display, lastFocusedClient->parent, activeBorderColor.pixel);
#ifdef SWMPANELTASKBAR
    showClients();
#endif//SWMPANELTASKBAR    
#endif//SHOWFOCUS

#ifdef MINIMAL
    XSetInputFocus(display, client->window, RevertToPointerRoot, CurrentTime);
#else//MINIMAL

//#ifdef NOMOUSEFOCUS
//    ignoreXErrors(XSetInputFocus(display, client->window, RevertToParent, CurrentTime));
//#else
    ignoreXErrors(XSetInputFocus(display, client->window, RevertToPointerRoot, CurrentTime));
//#endif//NOMOUSEFOCUS    
        
#endif//MINIMAL
}//focusClient

#ifndef MINIMAL
#ifndef PIXMAPS   
// -getClickedItem---------------------------------------------------------
//! finds out, which button was clicked
/*!
of this functions exist two versions, one window based for PIXMAPS 
mode and one simple for non-PIXMAPS mode, this is the non-pixmap version

\returns one of: TITLEBAR, CLOSEBOX, MAXBOX, STICKYBOX or WINDOWBODY
\param client 	the client clicked
\param x	x position of the click within the window
\param y	y position of the click within the window

\verbatim
   v1.3.0-pre1: initial release
   v1.3.0-pre2: added WINDOWBODY
                this function is still very poor but it may work for a while
   v1.3.0-pre3: redesigned, should work now as expected, two versions for
                pixmaps/nopixmaps		
   v1.3.0-pre4: no change
   v1.3.0-pre5: changed to return windowbody if frame is hidden
   v1.3.1:      removed in MINIMAL mode
\endverbatim
*/
int getClickedItem(Client *client, int x, int y)
{ int buttonNumber;

#ifdef MWM  
  if (client->frameState==FRAMESHOWN)
  { 
#endif//MWM
    if (y<=TITLEBARHEIGHT)
    { // titlebar was clicked

      buttonNumber=((client->width-x)/TITLEBARHEIGHT)+1;

      if (buttonNumber==CLOSEBOX) 
        return CLOSEBOX;

      if (buttonNumber==MAXBOX) 
#ifdef MAXIMIZE
        return MAXBOX;
#else//MAXIMIZE
#ifdef GNOME
        return STICKYBOX;
#endif//GNOME 	
#endif//MAXIMIZE 

#ifdef GNOME    
#ifdef MAXIMIZE
      if (buttonNumber==STICKYBOX) 
        return STICKYBOX;
#endif//MAXIMIZE  
#endif//GNOME 
      return TITLEBAR;
    }//if y<TITLEBARHEIGHT
    else    
    { return WINDOWBODY;
    }
#ifdef MWM    
  } 
#endif//MWM  
  return TITLEBAR;
}//getClickedItem - no PIXMAP 

#else // ------- PIXMAPS --------- 
//  getClickedItem
//! finds out, which button was clicked
/*!
of this functions exist two versions, one window based for PIXMAPS mode and one simple
for non-PIXMAPS mode, this is the pixmap version

\returns one of: TITLEBAR, CLOSEBOX, MAXBOX, STICKYBOX or WINDOWBODY
\param client 	the client clicked
\param window	clicked window inside the client
\param y	y position of the click within the window

\verbatim
   v1.3.0-pre1: initial release
   v1.3.0-pre2: added WINDOWBODY
                this function is still very poor but it may work for a while
   v1.3.0-pre3: redesigned, should work now as expected, two versions for
                pixmaps/nopixmaps		
   v1.3.0-pre4: no change
   v1.3.0-pre5: changed to return windowbody if frame is hidden
   v1.3.1:      removed in MINIMAL mode
\endverbatim
*/
int getClickedItem(Client *client, Window window, int y)
{
  if (window==client->closebox) return CLOSEBOX;

#ifdef MAXIMIZE
  if (window==client->maxbox)    return MAXBOX;
#endif//MAXIMIZE 
  
#ifdef GNOME
  if (window==client->stickybox) return STICKYBOX;
#endif//GNOME 

#ifdef MWM  
  if (window==client->parent) 
  {
    if (client->frameState==FRAMESHOWN)
    { if (y<=TITLEBARHEIGHT) return TITLEBAR;
      else return WINDOWBODY;
    }
    else
    { return WINDOWBODY;
    }  
  } else 
  { return WINDOWBODY;
  }
#endif//MWM 
  return TITLEBAR;
}// end getClickedItem - PIXMAPS 
#endif//PIXMAPS 
#endif//MINIMAL
	

// -changeGravity---------------------------------------------------
//! changes the window's gravity
/*!
\param client 		the client meant
\param multiplier	number of pixels to graviate the window

\verbatim
   v1.3.0-pre1: initial release
   v1.3.0-pre2: no change
   v1.3.0-pre3: no change
\endverbatim
*/
void changeGravity(Client *client, int multiplier)
{ int dy = 0;
  int gravity;

#ifdef TRACE
  trace (1,"changeGravity"); 
#endif//TRACE

  gravity = (client->size->flags & PWinGravity) ?
             client->size->win_gravity : NorthWestGravity;

  switch (gravity) 
  { case NorthWestGravity:
    case NorthEastGravity: 
    case NorthGravity:  
         dy = TITLEBARHEIGHT; 
         break;
#ifndef MINIMAL
    case CenterGravity: dy = TITLEBARHEIGHT/2; break;
#endif//MINIMAL    
  }//switch
  client->y += multiplier * dy;

#ifdef TRACE
  trace (-1,"changeGravity"); 
#endif//TRACE
}//changeGravity 


// -getWindowUnderPointer------------------------------------
//! returns the window unter the mouse cursor
/*!
\returns the window under the pointer
\param givenWin usually contains the root window

\verbatim
   v1.3.0-pre1: initial release
   v1.3.0-pre2: no change
   v1.3.0-pre3: added extra check
\endverbatim   
*/ 
#ifndef MINIMAL
Window getWindowUnderPointer(Window givenWin) 
{ 
  Window dummywin, returnWin; 
  int x, y, wx, wy, mask; 
 
#ifdef TRACE 
  trace (0,"getWindowUnderPointer(*)"); 
#endif//TRACE 
 
  if (givenWin) 
  { XQueryPointer(display, givenWin, &dummywin, &returnWin, 
                  &x, &y, &wx, &wy, &mask); 
   return returnWin; 
  }
  
#ifdef VERBOSE
  swmerr ("no window under cursor");
#endif//VERBOSE
  return 0; 
}// getWindowUnderPointer 
#endif//NOT MINIMAL 

// -sendXMessage----------------------------------------------------------
//! sends a xwindow event to the window "givenWin" 
/*!
\param givenWin	the window to which the event should be sent
\param xatom 	the message-type wrapped into a xatom
\param value	long value of the message

\verbatim
   v1.3.0-pre1: initial release
   v1.3.6:      now checking for givenWin in CHECKCLIENTS mode
\endverbatim
*/

int sendXMessage(Window givenWin, Atom xatom, long value)
{ XEvent event;

#ifdef TRACE
  trace (0,"sendXMessage(*)");
#endif//TRACE

#ifdef CHECKCLIENTS
  if (givenWin)
  {
#endif//CHECKCLIENTS

    event.type 			= ClientMessage;
    event.xclient.window 	= givenWin;
    event.xclient.message_type 	= xatom;
    event.xclient.format 	= 32;
    event.xclient.data.l[0] 	= value;
    event.xclient.data.l[1] 	= CurrentTime;

    return XSendEvent(display, givenWin, False, NoEventMask, &event);
#ifdef CHECKCLIENTS
  } 
#ifdef ERRORMSG  
  else
  {   swmerr ("ERROR: sendXMessage window: %i not valid",givenWin);
#ifdef DEBUGPARANOID
      exit(-1);
#endif//DEBUGRARANOID      
  }//givenwin
#endif//ERRORMSG  
#endif//CHECKCLIENTS  
}//sendXmessage 

// ***************************************************************************
